import time

import numpy as np
from pstl.scripts.janus.utls.agilent import scanDAQRESVDC
from pstl.scripts.janus.utls.agilent import scanDAQVDC
from pstl.scripts.janus.utls.langmuir import langmuir_calc_ps_grnd as lang_calc

def update(k,data,*args):
    # k==channel, aka subplot axes
    vprobe=data[0]
    current=data[1]
    # choose channel data to output
    x_out=vprobe[k]
    y_out=current[k]

    #update any out put args (none here)
    args_out=args
    out={
            'x_out':x_out,
            'y_out':y_out,
            'args_out':args_out
            } 
    return out


def scan(j,ps,daq,slot,channels,voltages,resistance,voltage_delay):
    # j==which iteration of _animate -> jj==which voltage
    lenv=len(voltages)
    jj=np.remainder(j,lenv)
    # then this round of voltage is
    v=voltages[jj]
    # set power supply voltage
    ps.setVoltage(v)
    # if voltage_delay
    if voltage_delay is not None:
        time.sleep(voltage_delay)
    # if resistance is True, probe resisitance
    # elif False, raise error
    # elif None, raise error
    # else, use this list as resistance values for
    if resistance is True:
        r,v=scanDAQRESVDC(daq,slot,channels)
    elif isinstance(resistance,(list,np.ndarray)) and \
            len(resistance)==len(channels)-1:
        r=resistance
        v=scanDAQVDC(daq,slot,channels)
    else: #resistance is False or resistance is None:
        s="'resistance' must be True or a list length of channels minus 1"
        raise ValueError(s)
    # seperate voltages 
    vps=v[0]
    vr=v[1:]

    # perfrom lang analysis
    vprobe,current=lang_calc(vps,vr,r)
    return vprobe,current
    



"""

Testing get_electron_temperature func - funcs_electon_temperature.py


"""

import numpy as np

from pstl.utls.verify import verify_type
from pstl.utls.functionfit.helpers import find_fit
from pstl.diagnostics.probes.langmuir.single.analysis.floating_potential import get_floating_potential

default_fit_kwargs = {
    'deg': 1, 'power': 1, 'polarity': 1,
    'reverse': False, 'return_best': True, 'fit_type': "exponential",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': None, 'fstep_type': None, 'fstep_adjust': True, 'fitmax': None,
    'bstep': None, 'bstep_type': None, 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.25,
    'threshold_rsq': 0.95, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 1.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}


def get_electron_temperature(*args, method=0, **kwargs):
    # Declare available methods
    available_methods = {
        0: 'fit',
    }

    # Converts method: str -> method: int if method is a str
    if isinstance(method, str):
        reversed_methods = {v: k for k, v in available_methods.items()}
        method = reversed_methods.get(method, None)

    # check for match and get which function to use
    # raises value error with options if failed to match
    if method == 0:  # default
        func = get_electron_temperature_fit
    else:  # makes a table of options if error occurs
        table = "\n".join([f"{k}\t{v}" for k, v in available_methods.items()])
        raise ValueError(
            f"Matching method not found: {method}\nChoose from one of the available options:\n{table}")

    # Call funtion and return result
    return func(*args, **kwargs)


def get_electron_temperature_fit(
        voltage, current, *args,
        V_f=None, V_s=None,
        **kwargs):

    # verify floating potential is a usable value if given else solve for floating potential
    if V_f is not None:  # V_f is given and evaluated at this location
        verify_type(V_f, (int, float, np.int64, np.float64, np.ndarray), 'V_f')
    else:
        # determine starting point (all positive after V_f)
        floating_kwargs = kwargs.pop('V_f_kwargs', {})
        floating_kwargs.setdefault('method', "consecutive")
        # get floating potential
        V_f, _ = get_floating_potential(
            voltage, current, **floating_kwargs)

    # Once floating Potential is found, find its index w.r.t. data
    istart = np.where(voltage < V_f)[0][-1]+1
    # Get data from positive current values (above floating)
    xdata = voltage[istart:]
    ydata = current[istart:]

    # Set defaults for find fit search
    # sets default values for electron retarding fit, then updates with passed kwargs, then updates with fit_kwargs
    fit_kwargs = dict(default_fit_kwargs)
    fit_kwargs.update(kwargs.pop('fit_kwargs', {}))

    # solve for fit
    elec_ret_fit = find_fit(xdata, ydata, **fit_kwargs)

    # get electron temperature from fit polynomial conversion
    KTeV = 1/elec_ret_fit.poly.convert().coef[-1]

    # to return format (value, extras: Dict[str,Any])
    other = {"fit": elec_ret_fit}
    return KTeV, other



import datetime
import time 
import atexit
from contextlib import redirect_stdout
import os
import inspect

import serial
import numpy as np

from pstl.tools.alerts import logs
from pstl.tools.alerts import center
from pstl.tools.alerts import monitor


def warning_send(s,attempt:str=None):
    Sender=center.sender('pstlrc','mynotifier')
    now = datetime.datetime.now()
    now_date=now.strftime("%b-%d-%Y")
    
    subject="Pressure Warning: "+now_date+attempt
    body=s
    msgtype="warning"

    groupname='rocket'

    center.contact_group(Sender,body,groupname,'all',subject=subject)

def update_send(s):
    Sender=center.sender('pstlrc','mynotifier')
    now = datetime.datetime.now()
    now_date=now.strftime("%b-%d-%Y")
    
    subject="Pressure Update: "+now_date
    body=s
    msgtype="info"

    groupname='rocket'

    center.contact_group(Sender,body,groupname,'email',subject=subject)

def gas_correction(gas,logname):
    gas=gas.upper()
    if gas == "air".upper():
        gcf=1
    elif gas == "n2".upper():
        gcf=1
    elif gas == "xe".upper():
        gcf=0.4
    else:
        sout=("%s is not a known gas\nUsing gcf=1.0\n"%(gas))
        print(sout)
        loggit(logname,sout)
        gcf=1.0

    return gcf


def bpg402_torr(hv,lv,gas="N2",logname=None):
    e=np.divide((np.multiply(hv,256)+lv),4000)-12.625
    p=np.power(10,e)
    if gas != "N2".upper() or gas != "air".upper():
        gcf=gas_correction(gas,logname)
        p=p*gcf
    return p

def find75(o):
    i7=[i7 for i7,x in enumerate(o) if x==7]
    i5=[i5 for i5,x in enumerate(o) if x==5]
    print(i7)
    print(i5)
    if len(i7)==1 and len(i5)==1:
        print("pass")
        out=o[i7[0]:]+o[:i7[0]]
        if i7[0]+1==i5[0]:
            ihv=int(i7[0]+4)
            ilv=int(i7[0]+5)
    else:
        print("fail")
                
    return out,ihv,ilv

def calculate(ser,gas="N2",resetmax=5,logname=None):
    gas=gas.upper()
    reset=0
    RESET=[]
    collect=True
    while collect:
        if reset<resetmax:
            out=get_bytes(ser)
            if out is str:
                print(out)
                error=out
            else:
                error=out[3]
            if error==0:
                collect=False
                # collect
                hv=out[4];lv=out[5]
                p=bpg402_torr(hv,lv,gas=gas,logname=logname)
            else:
                reset = reset+1
                RESET.append(out)
        else:
            now = datetime.datetime.now()
            now_date=now.strftime("%b-%d-%Y")
            now = now.strftime("%b-%d-%Y %H:%M:%S")
            p="Error: @ %s had %i resets in a row\n%s\n"%(now,resetmax,str(RESET))
            collect=False
    return p,reset,out

def check(p,pressureLimit,gas="N2"):
    if p>pressureLimit:
        ok=False
    else:
        ok=True
    return ok

def get_bytes(ser):
    o=list(ser.read(9))
    I7=[i7 for i7,x in enumerate(o) if x==7]
    I5=[i5 for i5,x in enumerate(o) if x==5]
    if I7 != []  and I5 != []:
        for k in I7:
            for z in I5:
                if k+1==z:
                    i7=k
                    i5=z
                    out=o[i7:]+o[:i7]
                elif k+1==9 and z==0:
                    i7=k
                    i5=z
                    out=o[i7:]+o[:i7]
                else:
                    now = datetime.datetime.now()
                    now_date=now.strftime("%b-%d-%Y")
                    now = now.strftime("%b-%d-%Y %H:%M:%S")
                    out="Error: @ %s could not find 7, 5 in correct order\n%s\n"%(now,str(o))
    else:
        now = datetime.datetime.now()
        now_date=now.strftime("%b-%d-%Y")
        now = now.strftime("%b-%d-%Y %H:%M:%S")
        out="Error: @ %s could not find 7, 5 in correct order\n%s\n"%(now,str(o))

    return out


def exit_handler(LogicReset,logfile):
        if LogicReset==0:
            pass
            sout=("Info: No Logic Resets")
        else:
            sout=("Info: Number of Logic Resets %i"%(LogicReset))
        print(sout)
        loggit(logfile,sout)
    
def setup_loggit(filename:str=None):

    now = datetime.datetime.now()
    now_date=now.strftime("%Y_%m_%d_%H_%M_%S")
    if filename is None:
        frame=inspect.stack()[1]
        module=inspect.getmodule(frame[0])
        filename=os.path.splitext(os.path.basename(frame.filename))[0]
        filename=('out_'+filename+'_'
                +now_date+'.txt')
    logs_path=inspect.getfile(logs)
    logs_path=os.path.dirname(logs_path)
    full_file_location=os.path.join(logs_path,filename)
    print("Log file saved to:\n%s\n"%(full_file_location))
    return full_file_location

def loggit(filename:str,output):
    with open(filename,'a') as f:
        f.write(output)
class Args():
    def __init__(self,*args):
        self.args=args
    def update(self,*args):
        self.args=args

def loop(args):
    (logname,now_start,loggit,repeatSend,firstSend,
        send,sendUpdate,delay_update,repeat_1,
        delay_1,repeat_2,delay_2,now_last_send,now_last_update,
        logicReset,pressureLimit,gas,ser)=args

    p,reset,out=calculate(ser,gas=gas,logname=logname)
    if reset != 0:
        logicReset=logicReset + 1
    if p is str:
        ok=False
        s=p
    else:
        ok=check(p,pressureLimit,gas=gas)
        if ok:
            now = datetime.datetime.now()
            now_date=now.strftime("%b-%d-%Y")
            now = now.strftime("%b-%d-%Y %H:%M:%S")
            s="Update: @ %s Pressure=%.3e torr using %s\nBytes: %s\n"%(now,p,gas,str(out))
        else:
            now = datetime.datetime.now()
            now_date=now.strftime("%b-%d-%Y")
            now = now.strftime("%b-%d-%Y %H:%M:%S")
            s="Warning: @ %s Pressure=%.3e torr using %s\nBytes: %s\n"%(now,p,gas,str(out))


    if ok:
        pass
    else:
        loggit(logname,s)
        print(s)
        attempt=None

        if now_last_send is None:
            now_last_send=time.time()

        if firstSend:
            sout='Sending First Notification\n'
            loggit(logname,sout)
            print(sout)
            attempt=' (First Attempt)'
            firstSend=False
        elif repeatSend:
            if repeat_1:
                if time.time()-now_last_send >= delay_1:
                    sout='Sending Second Notification\n'
                    loggit(logname,sout)
                    print(sout)
                    attempt=' (Second Attempt)'
                    send=True
                    repeat_1=False
                else:
                    send=False
            else:
                if repeat_2:
                    if time.time()-now_last_send >= delay_2:
                        sout='Sending Third Notification\n'
                        loggit(logname,sout)
                        print(sout)
                        attempt=' (Third Attempt)'
                        send=True
                        repeat_2=False
                    else:
                        send=False
        else:
            send=False

        if send:
            warning_send(s,attempt)
            send=False
        elif sendUpdate:
            if now_last_update is None:
                now_last_update=time.time()
            if time.time()-now_last_update >= delay_update:
                sout="Sending Update\n"
                print(sout)
                update_send(s)
                now_last_update=time.time()
        else:
            pass

    fargs=(logname,now_start,loggit,repeatSend,
            firstSend,send,sendUpdate,
            delay_update,repeat_1,delay_1,repeat_2,delay_2,
            now_last_send,now_last_update,
            logicReset,pressureLimit,gas,ser)

    return p,fargs


def main():
    logname=setup_loggit()
    now=datetime.datetime.now()
    now_start=now.strftime("%b-%d-%Y %H:%M:%S")
    loggit(logname,'Start of %s\n'%(now_start))
    repeatSend=True
    firstSend=True
    send=True
    sendUpdate=True
    delay_update=3600
    now_last_update=None
    repeat_1=True
    delay_1=20
    repeat_2=True
    delay_2=60
    now_last_send=None
    logicReset=0
    pressureLimit=1e-3
    gas="n2"
    ser=serial.Serial("COM1")
    # first grab is iffy
    # so ignore it
    out=get_bytes(ser)
    fargs=(logname,now_start,loggit,repeatSend,firstSend,send,sendUpdate,
            delay_update,repeat_1,delay_1,repeat_2,delay_2,now_last_send,
            now_last_update,logicReset,pressureLimit,gas,ser)
    inficon_pressure=monitor.Monitor(1,1,\
            xlabel='time [s]',ylabel='pressure [torr]',\
            title='Inficon BPG402S Pressure',alerts=False,\
            func=loop,fargs=(fargs),start_time=True,\
            ylimit_style='m',ylimit=[-1,1],logy=False
            )
    print("running ...")
    try:
        inficon_pressure.monitor()
    except KeyboardInterrupt:
        loggit(logname,'End of %s\n'%(now_start))
        print("exiting ...")
    exit_handler(logicReset,logname)


if __name__ == "__main__":
    main()



def cmdGetVoltageDC(loc):
    return "MEAS:VOLT:DC? (@%s)"%(loc)

def cmdGetVoltageAC(loc):
    return "MEAS:VOLT:AC? (@%s)"%(loc)

def cmdGetTemperatureTCK(loc):
    return "MEAS:TEMP? TC,K,(@%s)"%(loc)

def cmdGetResistance(loc):
    return "MEAS:RES? (@%s)"%(loc)



import copy
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib import pyplot as plt
from matplotlib.widgets import Button
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk  # type: ignore
import tkinter as tk
from tkinter import Frame

from pstl.utls.verify import verify_type
from pstl.utls.helpers import count_placeholders
from backend_matplotlib_interactive_cursor import AnnotatedCursor


class ButtonBaseBuilder:
    def __init__(self, ax, ax_plot_func=None):
        self.name = "Base Builder"
        self.ax = ax
        self._history = []
        self._originals = {'artists': len(ax.artists), 'collections': len(ax.collections), 'images': len(
            ax.images), 'lines': len(ax.lines), 'patches': len(ax.patches), 'texts': len(ax.texts)}

        # adds connetion to button pressed events
        self.cid = None
        #ax.figure.canvas.mpl_connect('button_press_event', self)

    def disconnect(self, val):
        self.ax.figure.canvas.mpl_disconnect(self.cid)

    def connect(self, val):
        self.cid = self.ax.figure.canvas.mpl_connect(
            'button_press_event', self)


class VerticalLineBuilder(ButtonBaseBuilder):
    def __init__(self, ax,
                 ax_plot_func=None,
                 control_panel=False, tk_home=None,
                 ax_plot_kwargs: dict = {}, axvline_kwargs: dict = {},
                 **kwargs):
        ButtonBaseBuilder.__init__(self, ax, ax_plot_func)
        self.name = "Vertical Line Builder"
        self.xs = []
        self.label_format = kwargs.get('label_format', None)
        self._original_number_of_plots = self._originals['lines']
        self._original_index = self._original_number_of_plots - 1
        self._xs_counter = 0
        self.ax_plot_func = ax_plot_func if ax_plot_func is not None else ax.plot
        self.ax_plot_kwargs = ax_plot_kwargs
        self.axvline_kwargs = axvline_kwargs
        self.control_panel = None
        self.tk_home = tk_home

        # If tk_home is a not None, then checks it is correct type, If pass:
        # makes a canvas to edit on tkinter frame/window
        if self.tk_home is not None:
            verify_type(self.tk_home, (tk.Frame, tk.Tk), 'tk_home')
            self.canvas = MatplotlibCanvasFrameTk(self.ax, self.tk_home)

        # if control_panel is True, make the control panel
        # if tk_home is a tk.Frame or tk.Tk object, then tkinter will be used, else matplotlib figure
        if control_panel:
            self._make_control_panel()

    def __call__(self, event):
        if event.inaxes != self.ax.axes:
            return
        self.add_vertical_line(event.xdata)

    def add_vertical_line(self, val):
        self.xs.append(val)
        axvline_kwargs = copy.deepcopy(self.axvline_kwargs)
        # add color default if not given per added line
        axvline_kwargs['color'] = axvline_kwargs.get(
            'color', "C"+str(self._xs_counter+self._original_number_of_plots))
        # check if label needs arguments
        label = axvline_kwargs.get('label', None)
        if label is not None:
            num_ph = count_placeholders(label)
            if num_ph != 0:
                label = label.format(val)
        axvline_kwargs['label'] = label
        self.ax.axvline(val, **axvline_kwargs)
        self.ax.figure.canvas.draw()
        self._xs_counter += 1

    def clear_lines(self):
        num_of_lines = len(self.ax.lines)
        max_index_of_lines = num_of_lines - 1
        for index in range(max_index_of_lines, self._original_index, -1):
            line = self.ax.lines[index]
            line.remove()
        self.xs = []
        self._xs_counter = 0
        self.ax.figure.canvas.draw()

    def undo(self):
        if len(self.xs) == 0:
            print("Nothing to undo")
        else:
            self.ax.lines[-1].remove()
            self._history.append(self.xs.pop())
            self._xs_counter -= 1
            self.ax.figure.canvas.draw()

    def redo(self):
        if len(self._history) == 0:
            print("Nothing to be redo")
        else:
            replot = self._history.pop()
            self.add_vertical_line(replot)

    def clear(self):
        self.clear_lines()

    def _make_control_panel(self):

        if self.tk_home is not None:  # tk control panel
            self.control_panel = VerticalLineControlPanelTk(
                self.tk_home, worker=self)  # type: ignore

        else:  # matplotlib control panel
            self.control_panel = plt.figure(
                FigureClass=VerticalLineControlPanelMatplotlib, worker=self)


class VerticalLineControlPanelMatplotlib(Figure):
    def __init__(self,  *args, worker=VerticalLineBuilder, **kwargs):
        Figure.__init__(self, *args, **kwargs)

        # make and place disconnect button
        btn_disconnect = Button(self.add_axes([0.1, 0.9, 0.15, 0.04]),
                                'disconnect', hovercolor='0.975')  # type: ignore
        btn_disconnect.on_clicked(worker.disconnect)
        # make and place connect button
        btn_connect = Button(self.add_axes([0.1, 0.8, 0.15, 0.04]),
                             'connect', hovercolor='0.975')  # type: ignore
        btn_connect.on_clicked(worker.connect)
        # make and place clear button
        btn_clear = Button(self.add_axes([0.1, 0.7, 0.15, 0.04]),
                           'clear', hovercolor='0.975')  # type: ignore
        btn_clear.on_clicked(worker.clear)

        self.buttons = [btn_disconnect, btn_connect, btn_clear]


class VerticalLineControlPanelTk(Frame):
    def __init__(self, root, *args, worker=VerticalLineBuilder, **kwargs):
        Frame.__init__(self, root, *args, **kwargs)

        # put Frame in root Frame
        self.grid(row=0, column=0)

        # make Label title
        title = tk.StringVar()
        title.set("Control Panel:\nVeritcal Line Builder")
        lbl_title = tk.Label(
            self, textvariable=title,
        )
        # make buttons
        btn_disconnect = tk.Button(
            self, text='Disconnect', command=lambda: worker.disconnect(None))  # type: ignore
        btn_connect = tk.Button(
            self, text='Connect', command=lambda: worker.connect(None))  # type: ignore
        btn_clear = tk.Button(self, text='Clear',  # type: ignore
                              command=lambda: worker.clear())  # type: ignore
        btn_undo = tk.Button(
            self, text="Undo", command=lambda: worker.undo())  # type: ignore
        btn_redo = tk.Button(
            self, text="Redo", command=lambda: worker.redo())  # type: ignore

        # pack label "Control Panel" into this Frame
        lbl_title.grid(row=0, column=0, columnspan=3, pady=10)
        # pack buttons into this frame
        btn_undo.grid(row=1, column=0, sticky="NSWE")
        btn_redo.grid(row=1, column=1, sticky="NSWE")
        btn_connect.grid(row=2, column=0, columnspan=2, sticky="NSWE")
        btn_disconnect.grid(row=3, column=0, columnspan=2, sticky="NSWE")
        btn_clear.grid(row=4, column=0, columnspan=2, sticky="NSWE")

        self.widgets = {'labels': {'title': {'object': lbl_title, 'var': title}},
                        'buttons': {'undo': btn_undo, 'redo': btn_redo,
                                    'connect': btn_connect, 'disconnect': btn_disconnect, 'clear': btn_clear},
                        'check_buttons': {}
                        }


class MatplotlibCanvasFrameTk(Frame):
    def __init__(self, worker, root, *args, **kwargs):
        Frame.__init__(self, root, *args, **kwargs)
        self.grid(row=0, column=1)
        self.ax = worker

        # Create the Matplotlib canvas
        self.canvas = FigureCanvasTkAgg(
            self.ax.figure, master=self)
        self.canvas.draw()
        self.widget = self.canvas.get_tk_widget()
        self.widget.pack(
            side=tk.TOP, fill=tk.BOTH, expand=1)
        # pack_toolbar=False will make it easier to use a layout manager later on.
        self.toolbar = NavigationToolbar2Tk(
            self.canvas, self, pack_toolbar=True)
        self.toolbar.update()


class FloatingPotentialBuilder(VerticalLineBuilder):
    def __init__(self, ax, ax_data=None, ax_plot_func=None, control_panel=False, tk_home=None,
                 ax_plot_kwargs: dict = {}, axvline_kwargs: dict = {}, annotated_cursor_kwargs: dict = {},
                 **kwargs):
        axvline_kwargs['label'] = axvline_kwargs.get(
            'label', r"$V_{{f}}$ = {0:.2f}V")
        VerticalLineBuilder.__init__(
            self, ax, ax_plot_func, control_panel, tk_home, ax_plot_kwargs, axvline_kwargs)
        self.ax_data = ax.lines[0] if ax_data is None else ax_data
        # default settings for annotated cursor kwargs
        annotated_cursor_kwargs['ax'] = annotated_cursor_kwargs.get('ax', ax)
        annotated_cursor_kwargs['useblit'] = annotated_cursor_kwargs.get(
            'useblit', True)
        annotated_cursor_kwargs['color'] = annotated_cursor_kwargs.get(
            'color', 'black')
        annotated_cursor_kwargs['linestyle'] = annotated_cursor_kwargs.get(
            'linestyle', '--')
        annotated_cursor_kwargs['linewidth'] = annotated_cursor_kwargs.get(
            'linewidth', 0.8)
        annotated_cursor_kwargs['numberformat'] = annotated_cursor_kwargs.get(
            'numberformat', "{0:.2f}\n{1:.2f}")
        self.cursor = AnnotatedCursor(self.ax_data, **annotated_cursor_kwargs)

        if control_panel:
            # add RadioButton for Legend
            self.control_panel.widgets['labels']['title']['var'].set(
                "Control Panel:\nFloating Potential")
            self._update_control_panel()

    def legend(self):  # , var: bool = True):
        var = self.control_panel.widgets['check_buttons']['legend']['var'].get(
        )
        if var:  # True
            self.ax.legend()
            self.ax.figure.canvas.draw()
        else:
            self.ax.get_legend().remove()
            self.ax.figure.canvas.draw()

    def _update_control_panel(self):
        frame = self.control_panel
        var = tk.BooleanVar()
        cbtn_legend = tk.Checkbutton(
            frame, text='Legend',
            variable=var,
            onvalue=True,
            offvalue=False,
            command=self.legend)
        self.control_panel.widgets['check_buttons']['legend'] = {
            'object': cbtn_legend, 'var': var}
        cbtn_legend.grid(row=5, column=0, columnspan=2, sticky="NWSE")



from alicat import FlowController as FC


K=FC(port="COM5",address='K')
M=FC(port="COM5",address="M")
N=FC(port="COM5",address="N")

MFCs=[K,M,N]

# set gas type
for k in MFCs:
    k.set_gas('N2')

# set flow rates
for k in MFCs:
    sccm=0

    k.set_flow_rate(sccm)

# view results
for k in MFCs:
    print(k.address)
    print(k.get())



import matplotlib.pyplot as plt
import matplotlib.animation as animation
import time
import datetime
import math


class Subplot():
    def __init__(self,ax,*args,**kwargs):
        self.ax=ax
        self.x=[]
        self.y=[]

        self.start_time=kwargs.get('start_time',None)
        self.logx=kwargs.get('logx',False)
        self.logy=kwargs.get('logy',False)
        self.xlimit_style=kwargs.get('xlimit_style',None)
        self.xlimit=kwargs.get('xlimit',None)
        self.ylimit_style=kwargs.get('ylimit_style',None)
        self.ylimit=kwargs.get('ylimit',None)


        self.func=kwargs.get('func',None)
        self.fargs=kwargs.get('fargs',None)

        self.location=kwargs.get('location',None)
        self.xlabel=kwargs.get('xlabel',None)
        self.ylabel=kwargs.get('ylabel',None)
        self.title=kwargs.get('title',None)

class Monitor():
    def __init__(self,nrows:int=1,ncols:int=1,**kwargs):
        self.nrows=nrows
        self.ncols=ncols

        self.alerts=kwargs.get('alerts',False)
        self.full_screen=kwargs.get('full_screen',False)

        fig,ax_list=\
                plt.subplots(nrows=nrows,ncols=ncols)
        self.figure=fig

        if nrows==1 and ncols==1:
            ax=[ax_list]
            nax=1
        else:
            ax=ax_list.ravel()
            nax=len(ax)
        self.axes=ax

        self.nax=nax
        subplot=[None]*nax
        for k in range(nax):
            subplot[k]=Subplot(ax[k],**kwargs)
        self.subplot=subplot

    # This function is called periodically from FuncAnimation
    def _animate(self, i):

        for k in range(self.nax):
            # set xs,ys
            xs=self.subplot[k].x
            ys=self.subplot[k].y

            # axes to use
            ax=self.axes[k]

            # Read temperature 
            if self.subplot[k].fargs is not None:
                y_out,self.subplot[k].fargs = \
                    self.subplot[k].func(self.subplot[k].fargs)
            else:
                y_out=self.sublpot[k].func()
            if self.subplot[k].start_time is True:
                x_out=datetime.datetime.now()
                self.subplot[k].start_time=x_out
            elif self.subplot[k].start_time is None:
                x_out=datetime.datetime.now()
            else:
                x_out=datetime.datetime.now()-self.subplot[k].start_time

            #roundoff museconds

            try:
                x_out=x_out.strftime('%H:%M:%S.%f')
            except:
                x_out=str(x_out-datetime.timedelta(microseconds=x_out.microseconds))


            # Add x and y to lists
            xs.append(x_out)
            ys.append(y_out)

            # Limit x and y lists to 20 items
            xs = xs[-20:]
            ys = ys[-20:]

            # Draw x and y lists
            ax.clear()
            if self.subplot[k].logx and self.subplot[k].logy:
                ax.loglog(xs, ys,\
                    marker='o',linestyle='-',markerfacecolor='none')
            elif self.subplot[k].logx:
                ax.semilogx(xs, ys,\
                    marker='o',linestyle='-',markerfacecolor='none')
            elif self.subplot[k].logy:
                ax.semilogy(xs, ys,\
                    marker='o',linestyle='-',markerfacecolor='none')
            else:
                ax.plot(xs, ys,\
                    marker='o',linestyle='-',markerfacecolor='none')

            self.subplot[k].x=xs
            self.subplot[k].y=ys

            # Format plot
            ax.grid(True)
            ax.tick_params(axis='x',labelrotation=45)
            #ax.set_xticklabls(ax.get_xticks(),rotation=45)
            ax.set(title=self.subplot[k].title)
            ax.set(xlabel=self.subplot[k].xlabel)
            ax.set(ylabel=self.subplot[k].ylabel)
            if self.subplot[k].ylimit_style == 'magnitude' or\
                    self.subplot[k].ylimit_style == 'm':
                        if self.subplot[k].ylimit is not None:
                            y_max=max(ys);y_min=min(ys)
                            magnitude_max=math.floor(math.log(y_max,10))
                            magnitude_min=math.floor(math.log(y_min,10))
                            #if self.subplot[k].ylimit is list\
                            #        and len(self.subplot[k].ylimit)==2:
                            if len(self.subplot[k].ylimit)==2:
                                        ylimit_max=math.pow(10,
                                                magnitude_max+\
                                                self.subplot[k].ylimit[1]
                                                )
                                        ylimit_min=math.pow(10,
                                                magnitude_min+\
                                                self.subplot[k].ylimit[0]
                                                )
                            elif self.subplot[k].ylimit is list\
                                    and len(self.subplot[k].ylimit)==1:
                                        ylimit_max=math.pow(10,
                                                magnitude_max+\
                                                self.subplot[k].ylimit[0]
                                                )
                                        ylimit_min=math.pow(10,
                                                magnitude_min-\
                                                self.subplot[k].ylimit[0])
                            elif self.subplot[k].ylimit is int\
                                    or self.subplot[k].ylimit is float:
                                        ylimit_max=math.pow(10,
                                                magnitude_max+\
                                                self.subplot[k].ylimit
                                                )
                                        ylimit_min=math.pow(10,
                                                magnitude_min-\
                                                self.subplot[k].ylimit)
                            else:
                                error_msg="\nERROR: '%s' is not a list, float or int\nit is a %s"%(str(self.subplot[k].ylimit),str(type(self.subplot[k].ylimit))) 
                                print(error_msg)
                            ax.set(ylim=[ylimit_min,ylimit_max])

                            

    def monitor(self,interval:float=1000):
        
        # Set up plot to call animate() function periodically
        ani = animation.FuncAnimation(self.figure, \
                self._animate, fargs=None, \
                interval=interval)
        wm = plt.get_current_fig_manager()
        if self.full_screen:
            wm.window.state('zoomed')
        plt.show()






"""

Testing get_ion_saturation current func - funcs_ion_saturation_current.py


"""
from typing import Optional, Union, Tuple, Dict, Any

import numpy as np
import numpy.typing as npt

from pstl.utls.verify import verify_type, verify_pair_of_1D_arrays
from pstl.utls.functionfit.helpers import find_fit, FunctionFit

from pstl.diagnostics.probes.langmuir.single.analysis.floating_potential import get_floating_potential, check_for_floating_potential, get_below_floating_potential
from pstl.diagnostics.probes.langmuir.single.analysis.plasma_potential import get_plasma_potential
from pstl.diagnostics.probes.langmuir.single.analysis.ion_current import check_for_ion_current_fit


default_fit_thin_kwargs = {
    'deg': 1, 'power': 1, 'polarity': 1,
    'reverse': False, 'return_best': True, 'fit_type': "linear",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': None, 'fstep_type': None, 'fstep_adjust': True, 'fitmax': None,
    'bstep': None, 'bstep_type': None, 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.30,
    'threshold_rsq': 0.95, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 1.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}

default_fit_thick_kwargs = {
    'deg': 1, 'power': 0.5, 'polarity': -1,
    'reverse': False, 'return_best': True, 'fit_type': "linear",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': None, 'fstep_type': None, 'fstep_adjust': True, 'fitmax': None,
    'bstep': None, 'bstep_type': None, 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.30,
    'threshold_rsq': 0.95, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 1.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}


def get_ion_saturation_current(
        *args,
        method: int | str = 0,
        **kwargs) -> Tuple[float, Dict[str, Any]]:
    """
    Parameters
    ----------
    method: str, optional
        if 'min' or 'minimum', then returns the the lowest current
        if 'vf' or 'floating', then returns the value at floating 
        potential voltage from a linear fit line that matches the 
        specified r^2, default r^2=0.95
    """
    # Declare available methods
    available_methods = {
        0: 'plasma',
        1: 'floating',
        2: 'minimum',
    }

    # Converts method: str -> method: int if method is a str
    if isinstance(method, str):
        reversed_methods = {v: k for k, v in available_methods.items()}
        # if method:str is not in avialiable_methods,
        # it will error out in check as it is set to the default and will not match
        method = reversed_methods.get(method, method)

    # check for match and get which function to use
    # raises value error with options if failed to match
    if method == 0:  # default
        func = plasma_potential_method
    elif method == 1:
        func = floating_potential_method
    elif method == 2:
        func = minimum_ion_current_method
    else:  # makes a table of options if error occurs
        table = "\n".join([f"{k}\t{v}" for k, v in available_methods.items()])
        raise ValueError(
            f"Matching method not found: {method}\nChoose from one of the available options:\n{table}")

    # Call funtion and return result
    return func(*args, **kwargs)


def get_ion_saturation_current_density(area, *args, method=0, I_is=None, **kwargs):
    if I_is is None:
        value, other = get_ion_saturation_current(
            *args, method=method, **kwargs)
        I_is = value
    else:
        other = {}
    J_is = np.divide(I_is, area)
    return J_is, other


def plasma_potential_method(
        voltage, current, *args,
        V_s: float | None = None, return_V_s: bool = False, I_es: float | None = None,
        **kwargs) -> Tuple[float, Dict[str, Any]]:
    # check if a plasma point is given, if non
    verify_pair_of_1D_arrays(voltage, current)
    if len(voltage) != len(current):
        raise ValueError(
            "Voltage and current arrays must have the same length.")

    # number of data points
    n = len(voltage)

    j = 0
    # Find the first transition point from negative to positive or zero current
    while j < n and current[j] < 0:
        j += 1

    # set some defaults for routine
    fit_kwargs = dict(default_fit_thin_kwargs)
    fit_kwargs.update(kwargs.pop('fit_kwargs', {}))
    # thus j now is where current transitions from negative to postive
    # loop through the points fitting a line
    # fit = routine_linear_fit(voltage[0:j], current[0:j], **kwargs)
    fit = find_fit(voltage[0:j], current[0:j], **fit_kwargs)

    # solve for ion saturation current
    if V_s is not None:  # vs is given and evaluated at this location
        verify_type(V_s, (int, float, np.int64, np.float64, np.ndarray), 'V_s')
    else:
        # if vs is none get plasma potential via
        # get_plasma_potential_consecutive of funcs_plasma_potential
        plasma_kwargs = kwargs.get('plasma_kwargs', {})
        V_s, _ = get_plasma_potential(
            voltage, current, **plasma_kwargs)

    # use plasma potential to find saturation current
    ion_saturation_current = fit(V_s)

    # if nan or positive
    if ion_saturation_current > 0 or np.isnan(ion_saturation_current):
        if I_es is None:
            raise NotImplementedError
        ion_saturation_current = calculate_ion_saturation_current_xenon(I_es)

    # Returns in formate (value, Dict[str,Any])
    other: dict[str, Any] = {"fit": fit}
    if return_V_s is True:
        other['V_s'] = V_s
    return ion_saturation_current, other


def floating_potential_method(
        voltage, current, *args,
        I_i_fit=None, I_i_method=None,
        V_f: float | None = None, return_V_f: bool = False,
        **kwargs) -> Tuple[float, Dict[str, Any]]:
    # check if a floating point is given, if non
    # verify arrays length and shape
    verify_pair_of_1D_arrays(voltage, current)
    if len(voltage) != len(current):
        raise ValueError(
            "Voltage and current arrays must have the same length.")

    # check  and setup xdata,ydata based on floaing potential
    V_f = check_for_floating_potential(V_f, voltage, current, **kwargs)
    iend, xdata, ydata = get_below_floating_potential(V_f, voltage, current)

    # check if fit needs to be found to evaluate at floating potential
    shape = kwargs.get("shape", None)
    fit_kwargs = kwargs.pop("fit_kwargs", {})
    I_i_fit = check_for_ion_current_fit(
        I_i_fit, voltage, current, shape, method=I_i_method, fit_kwargs=fit_kwargs)

    # use V_f to find saturation current
    ion_saturation_current = I_i_fit(V_f)

    # Returns in formate (value, Dict[str,Any])
    other: dict[str, Any] = {"fit": I_i_fit}
    if return_V_f is True:
        other['V_f'] = V_f
    return ion_saturation_current, other


def _floating_potential_method(
        voltage, current, *args,
        V_f: float | None = None, return_V_f: bool = False,
        **kwargs) -> Tuple[float, Dict[str, Any]]:
    # check if a floating point is given, if non
    # verify arrays length and shape
    verify_pair_of_1D_arrays(voltage, current)
    if len(voltage) != len(current):
        raise ValueError(
            "Voltage and current arrays must have the same length.")

    # number of data points
    n = len(voltage)

    j = 0
    # Find the first transition point from negative to positive or zero current
    while j < n and current[j] < 0:
        j += 1

    # thus j now is where current transitions from negative to postive
    # loop through the points fitting a line
    # fit = routine_linear_fit(voltage[0:j], current[0:j], **kwargs)
    fit = find_fit(voltage[0:j], current[0:j], **kwargs)

    # determine starting point (all positive after vf) if not given
    if V_f is None:
        # determine starting point (all positive after V_f)
        floating_kwargs = kwargs.pop('V_f_kwargs', {})
        floating_kwargs.setdefault('method', "consecutive")
        # get floating potential
        V_f, _ = get_floating_potential(
            voltage, current, **floating_kwargs)
    # verify V_f
    verify_type(V_f, (int, float, np.int64, np.float64, np.ndarray), 'V_f')

    # use V_f to find saturation current
    ion_saturation_current = fit(V_f)

    # Returns in formate (value, Dict[str,Any])
    other: dict[str, Any] = {"fit": fit}
    if return_V_f is True:
        other['V_f'] = V_f
    return ion_saturation_current, other


def minimum_ion_current_method(
        voltage: npt.ArrayLike,
        current: npt.ArrayLike,
        *args, **kwargs) -> Tuple[float, Dict[str, Any]]:
    # maybe later add checks for:
    # 1) increasing current with increasing voltage (aka neg current to pos current)
    # 2) vectors same length
    # 3) no odd things in ion saturation region such as duplicates or spikes

    # Following set format of (value, Dict)
    return np.min(current), {}


def calculate_ion_saturation_current_xenon(electron_sat) -> float:
    ion_sat = -np.divide(electron_sat, 323)
    if ion_sat is np.nan:
        raise ValueError
    return ion_sat



"""

Testing get_plasma_potential func - funcs_plasma_potential.py


"""
from typing import Tuple, Dict, Any

import numpy as np
import numpy.typing as npt
from numpy.polynomial import Polynomial as P

from pstl.utls.verify import verify_type, verify_array_polarity
from pstl.utls.functionfit.helpers import _fit_exponential_func, find_fit, FunctionFit
from pstl.diagnostics.probes.langmuir.single.analysis.floating_potential import get_floating_potential
from pstl.diagnostics.probes.langmuir.single.analysis.electron_temperaure import get_electron_temperature
from pstl.diagnostics.probes.langmuir.single.analysis.electron_saturation_current import get_electron_saturation_current


def find_plasma_potential(*args, method=0, **kwargs) -> float:
    value, _ = get_floating_potential(*args, method=method, **kwargs)
    return value


def get_plasma_potential(*args, method=0, **kwargs) -> Tuple[float, Dict[str, Any]]:
    # Declare available method
    available_methods = {
        0: 'intersection',
    }

    # Converts method: str -> method: int if method is a str
    if isinstance(method, str):
        reversed_methods = {v: k for k, v in available_methods.items()}
        method = reversed_methods.get(method, None)

    # check for match and get which function to use
    # raises value error with options if failed to match
    if method == 0:  # default
        func = intersection_method
    else:  # makes a table of options if error occurs
        table = "\n".join([f"{k}\t{v}" for k, v in available_methods.items()])
        raise ValueError(
            f"Matching method not found: {method}\nChoose from one of the available options:\n{table}")

    # Call funtion and return result
    return func(*args, **kwargs)


def intersection_method(
        voltage, current, *args,
        V_f=None, elec_ret_poly=None, elec_sat_poly=None,
        **kwargs) -> Tuple[float, Dict[str, Any]]:

    # determine starting point (all positive after vf) if not given
    if V_f is None:
        # determine starting point (all positive after V_f)
        floating_kwargs = kwargs.pop('V_f_kwargs', {})
        floating_kwargs.setdefault('method', "consecutive")
        # get floating potential
        V_f, _ = get_floating_potential(
            voltage, current, **floating_kwargs)

    # make sure have Electron Retarding Region Polynomial of type ~np.polynomial.Polynomial
    if elec_ret_poly is None:
        KT_e_kwargs = kwargs.get("KT_e_kwargs", {})
        # set defaults for getting temperature
        KT_e_kwargs.setdefault('V_f', V_f)
        KT_e_kwargs.setdefault('V_s', None)
        KT_e_kwargs.setdefault('method', 'fit')
        # overide fit kwargs defaults for electron temperaure fit (exponential)
        KT_e_kwargs.setdefault('fit_kwargs',
                               {
                                   'full': False,
                               }
                               )
        _, KT_e_others = get_electron_temperature(
            voltage, current, **KT_e_kwargs)
        elec_ret_poly = KT_e_others["fit"].poly.convert()
    verify_type(elec_ret_poly, P)

    # make sure have Electron Saturation Region Polynomial of type ~np.polynomial.Polynomial
    if elec_sat_poly is None:
        I_es_kwargs = kwargs.get("I_es_kwargs", {})
        # set defaults for getting temperature
        I_es_kwargs.setdefault('V_f', V_f)
        I_es_kwargs.setdefault('V_s', None)
        I_es_kwargs.setdefault('method', 'fit')
        I_es_kwargs.setdefault('elec_ret_poly', elec_ret_poly)
        # overide fit kwargs defaults for electron temperaure fit (exponential)
        I_es_kwargs.setdefault('fit_kwargs',
                               {
                                   'full': False,
                               }
                               )
        _, I_es_others = get_electron_saturation_current(
            voltage, current, **I_es_kwargs)
        elec_sat_poly = I_es_others["fit"].poly.convert()
    verify_type(elec_sat_poly, P)

    # find intersection
    # intersection = find_intersection(elec_sat_fit, elec_sat_fit)
    p = elec_sat_poly-elec_ret_poly
    intersection = p.roots()

    # checks that instersection is made
    if intersection is None:
        raise ValueError("Intersection not found")

    # gets the plasma potential at intersection
    plasma_potential = intersection[0]

    # prepare returns
    others = {"method": "intersection"}
    # returns (value, other: Dict[str,Any])
    return plasma_potential, others




from pstl.instruments.daq.agilent import channel

def loop_setup(channels,chtype,slot=None):
    nchannels = int(len(channels)-1)
    bool_list = isinstance(chtype,list)
    if bool_list:
        if len(chtype)==nchannels:
            # do this if same len
            __loop_multi(channels,chtype,slot)
        elif len(chtype)==1:
            # only one type make all that type
            __loop_single(channels,chtype,slot)
        else:
            print("Length of chtype is less than nchannels."\
                    +"Configuring first %i"%(len(chtype)))
            __loop_multi(channels,chtype,slot)
    else:
        try:
            bool_str = isinstance(chtype,str)
            # only one type make all that type
            __loop_single(channels,chtype,slot)
        except TypeError:
            print("Entered '%s' is not valid"%(str(chtype)))
    return channels



def __loop_single(channels,chtype,slot=None):
    """
    Changes the each channel to chtype
    """
    for k in range(1,len(channels)):
        try:
            loc = int(slot*100 + k)
        except:
            loc = int(k)
        channels[k] = channel.CHANNEL(loc,chtype)
    return channels

def __loop_multi(channels,chtypes,slot=None):
    """
    Changes the each channel to its corresponding chtype
    """
    for k in range(1,len(chtype)+1):
        try:
            loc = int(slot*100 + k)
        except:
            loc = int(k)
        chtype = chtypes[k-1]
        channels[k] = channel.CHANNEL(loc,chtype)
    return channels



from pstl.utls import constants as c


class Plasma:
    def __init__(self, m_i, m_e=c.m_e, gas=None, *args, **kwargs) -> None:
        self._m_i = m_i
        self._m_e = m_e
        self._gas = gas

    @property
    def m_i(self):
        return self._m_i

    @property
    def m_e(self):
        return self._m_e

    @property
    def gas(self):
        return self._gas


class XenonPlasma(Plasma):
    def __init__(self, m_i=None, m_e=c.m_e, gas=None, *args, **kwargs) -> None:
        if m_i is None:
            m_i = 131.29*c.m_p  # amu*kg -> kg
        super().__init__(m_i, m_e, gas, *args, **kwargs)


class ArgonPlasma(Plasma):
    def __init__(self, m_i=None, m_e=c.m_e, gas=None, *args, **kwargs) -> None:
        if m_i is None:
            m_i = 39.948*c.m_p  # amu*kg -> kg
        super().__init__(m_i, m_e, gas, *args, **kwargs)



import numpy as np


# some functions are dependent on classes defined elsewhere
# here they are passed in and functions are called


def offVoltage(ps):
    ps.setVoltage(0)


def setupSupplyVoltages():
    vstart=-1
    vstop=1
    dv=1
    return np.arange(vstart,vstop+dv,dv)





from abc import ABC, abstractmethod, abstractproperty
from typing import Dict

import numpy as np
import pandas as pd

from pstl.utls.plasmas import Plasma
from pstl.utls.preprocessing import preprocessing_filter, preprocess_filter, smooth_filter
from pstl.diagnostics.probes.langmuir.single import SingleProbeLangmuir

from pstl.diagnostics.probes.langmuir.single.analysis.algorithm import topham, lobbia
from pstl.diagnostics.probes.langmuir.single.analysis.floating_potential import get_floating_potential
from pstl.diagnostics.probes.langmuir.single.analysis.ion_saturation_current import get_ion_saturation_current, get_ion_saturation_current_density
from pstl.diagnostics.probes.langmuir.single.analysis.electron_temperaure import get_electron_temperature
from pstl.diagnostics.probes.langmuir.single.analysis.electron_saturation_current import get_electron_saturation_current, get_electron_saturation_current_density
from pstl.diagnostics.probes.langmuir.single.analysis.plasma_potential import get_plasma_potential
from pstl.diagnostics.probes.langmuir.single.analysis.electron_density import get_electron_density
from pstl.diagnostics.probes.langmuir.single.analysis.ion_density import get_ion_density
from pstl.utls.constants import lambda_D
from pstl.utls.helpers import probe_radius_to_debye_length as get_sheath_type
from pstl.utls.helpers import method_function_combiner, method_selector
from pstl.utls.decorators import absorb_extra_args_kwargs, add_empty_dict
get_lambda_D = add_empty_dict(absorb_extra_args_kwargs(lambda_D))
get_sheath_type = add_empty_dict(absorb_extra_args_kwargs(get_sheath_type))

# maybe a concrete class called ProbeData that would jsut be raw_data, deleted_data, filtered_data, smoothed_data


class DiagnosticData:
    def __init__(self, data, deleted_data=None, filtered_data=None, smoothed_data=None, *args, **kwargs):
        # set different data versions
        self.set_data(
            data,
            deleted_data=deleted_data,
            filtered_data=filtered_data,
            smoothed_data=smoothed_data,
        )

    def set_data(self, raw_data, source=0, deleted_data=None, filtered_data=None, smoothed_data=None):
        self._raw_data = raw_data
        self._deleted_data = deleted_data
        self._filtered_data = filtered_data
        self._smoothed_data = smoothed_data

        # available sources
        available_sources = {
            0: "best",
            1: "raw_data",
            2: "filtered_data",
            3: "smoothed_data",
        }

        # Converts method: str -> method: int if method is a str
        if isinstance(source, str):
            reversed_sources = {v: k for k, v in available_sources.items()}
            source = reversed_sources.get(source, None)

        # choose data source
        if source == 0:
            if smoothed_data is True:
                # get smoothed of filtered data
                raise NotImplementedError
                data = smoothed_data
            elif isinstance(smoothed_data, pd.DataFrame):
                data = smoothed_data
            elif filtered_data is True:
                # get filtered data
                raise NotImplementedError
                data = filtered_data
            elif isinstance(filtered_data, pd.DataFrame):
                data = filtered_data
            else:
                data = raw_data

        elif source == 1:
            data = raw_data
        elif source == 2:
            if filtered_data is True:
                filtered_data = pd.DataFrame
            elif filtered_data is False or filtered_data is None:
                filtered_data = raw_data
            elif isinstance(filtered_data, pd.DataFrame):
                filtered_data = filtered_data
            else:
                raise TypeError(
                    "'filtered_data' can only type pd.DataFrame, bool, None: ", type(filtered_data))
            data = filtered_data
        elif source == 3:   # tecniquely filtered and and smoothed if filtered is True and smoothed is True
            if smoothed_data is True:
                smoothed_data = pd.DataFrame
            elif smoothed_data is False or smoothed_data is None:
                smoothed_data = raw_data
            elif isinstance(smoothed_data, pd.DataFrame):
                smoothed_data = smoothed_data
            else:
                raise TypeError(
                    "'smoothed_data' can only type pd.DataFrame, bool, None: ", type(smoothed_data))
            data = smoothed_data
        else:  # makes a table of options if error occurs
            table = "\n".join(
                [f"{k}\t{v}" for k, v in available_sources.items()])
            raise ValueError(
                f"Matching source not found: {source}\nChoose from one of the available options:\n{table}")

        # based on data argument set data variable to use
        self._data = data

    def preprocess(self, *args, source=0, **kwargs):
        filtered_data, deleted_data = preprocess_filter(
            self._data, *args, **kwargs)
        self.set_data(self._data, source=source,
                      deleted_data=deleted_data, filtered_data=filtered_data)

    def smooth(self, *args, source=0, **kwargs):
        smoothed_data = smooth_filter(self._data, *args, **kwargs)
        self.set_data(self._data, source=source, smoothed_data=smoothed_data)


class PlasmaProbeSolver(DiagnosticData, ABC):
    def __init__(self, Plasma: Plasma, Probe, Data: pd.DataFrame,
                 methods: Dict = {}, properties: Dict = {},
                 deleted_data: pd.DataFrame | bool | None = None,
                 filtered_data: pd.DataFrame | bool | None = None,
                 smoothed_data: pd.DataFrame | bool | None = None,
                 *args, **kwargs) -> None:
        # super(ABC, self).__init__()
        DiagnosticData.__init__(
            self, Data,
            deleted_data=deleted_data, filtered_data=filtered_data, smoothed_data=smoothed_data)

        # add verification here
        self._plasma = Plasma
        self._probe = Probe

        # create results dictionary
        self._results = self.set_available_plasma_properties(properties)

        # set defeult methods (abstract)
        self._methods = self.set_default_methods(methods)

    @property
    def plasma(self):
        return self._plasma

    @property
    def probe(self):
        return self._probe

    @property
    def data(self):
        return self._data

    @property
    def raw_data(self):
        return self._raw_data

    @property
    def deleted_data(self):
        return self._deleted_data

    @property
    def results(self):
        return self._results

    @property
    def methods(self):
        return self._methods

    @abstractmethod
    def set_default_methods(self, methods: Dict) -> Dict:
        pass
        # return {}

    @abstractmethod
    def set_available_plasma_properties(self, properties: Dict) -> Dict:
        pass


class PlasmaLangmuirProbeSolver(PlasmaProbeSolver):
    def __init__(self, Plasma: Plasma, Probe, Data, *args, **kwargs) -> None:
        super().__init__(Plasma, Probe, Data, *args, **kwargs)

        self._data_e = self.data.copy()

    def update_current_e(self, curret_i):
        self.data_e.current = np.subtract(self.data.current, curret_i)

    @property
    def data_e(self):
        return self._data_e


class SingleLangmuirProbeSolver(PlasmaLangmuirProbeSolver):
    def __init__(self, Plasma: Plasma, Probe: SingleProbeLangmuir, Data: pd.DataFrame,
                 methods: Dict = {}, properties: Dict = {},
                 *args, **kwargs) -> None:
        super().__init__(Plasma, Probe, Data,
                         methods=methods, properties=properties,
                         *args, **kwargs)

    def set_available_plasma_properties(self, properties: Dict) -> Dict:
        super().set_available_plasma_properties(properties)
        available_plasma_properties = {
            "V_f": {'value': None, 'other': None},
            "V_s": {'value': None, 'other': None},
            "KT_e": {'value': None, 'other': None},
            "n_e": {'value': None, 'other': None},
            "I_es": {'value': None, 'other': None},
            "J_es": {'value': None, 'other': None},
            "n_i": {'value': None, 'other': None},
            "I_is": {'value': None, 'other': None},
            "J_is": {'value': None, 'other': None},
            "lambda_De": {'value': None, 'other': None},
            "sheath": {'value': None, 'other': None},
        }
        return available_plasma_properties

    def set_default_methods(self, methods: Dict):
        # methods = super().setdefault_methods(methods)
        super().set_default_methods(methods)
        default_methods = {
            "V_f": 0,
            "V_s": 0,
            "I_is": 0,
            "n_i": 0,
            "I_es": 0,
            "KT_e": 0,
            "n_e": 0,
            "J_es": 0,
            "J_is": 0,
            "lambda_De": 0,
            "sheath": 0,
        }
        default_methods.update(methods)
        return default_methods

    def find_plasma_properties(self, algo=0, methods={}, *args, **kwargs):
        # algo is the default algo to use i.e. topham, lobia, etc.
        # verify methods is a dictionary
        # methods should not be modified unless you want to modify which methods are being used during the alogrythm.
        if not isinstance(methods, dict):
            raise ValueError(
                "'methods' must be a dictionary not: ", type(methods))

        # overwrite methods if passed in
        methods_to_use = dict(self.methods)
        methods_to_use.update(methods)

        # loop through plasma properties to get and perform each in given order
        # HARD CODEDIN
        func = method_selector(algo, available_algorithms,
                               available_algorithm_functions)
        data, results = func(
            self.data.voltage, self.data.current,
            self.probe.shape, self.probe.radius, self.probe.area,
            self.plasma.m_i, m_e=self.plasma.m_e,
        )
        self._results = results
        self._data = data

    def find_plasma_property(self, key, method=None, **kwargs):
        # choose which property
        if key == "V_f":
            func = get_floating_potential
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault("current", self.data.iloc[:, 1])  # type: ignore
        elif key == "V_s":
            func = get_plasma_potential
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault(
                "current", self.data_e.iloc[:, 1])  # type: ignore
            elec_ret_poly = self.results["KT_e"]["other"]["fit"].poly.convert(
            ) if "fit" in self.results["KT_e"]["other"] else None
            elec_sat_poly = self.results["I_es"]["other"]["fit"].poly.convert(
            ) if "fit" in self.results["I_es"]["other"] else None
            kwargs.setdefault("elec_ret_poly", elec_ret_poly)
            kwargs.setdefault("elec_sat_poly", elec_sat_poly)
        elif key == "I_is":
            func = get_ion_saturation_current
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault("current", self.data.iloc[:, 1])  # type: ignore
        elif key == "n_i":
            func = get_ion_density
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault("current", self.data.iloc[:, 1])  # type: ignore
            kwargs.setdefault("area", self.probe.area)
            kwargs.setdefault("KT_e", self.results["KT_e"]["value"])
            kwargs.setdefault("I_is", self.results["I_is"]["value"])
            kwargs.setdefault("m_i", self.plasma.m_i)
            kwargs.setdefault("m_e", self.plasma.m_e)
        elif key == "I_es":
            func = get_electron_saturation_current
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault(
                "current", self.data_e.iloc[:, 1])  # type: ignore
        elif key == "KT_e":
            func = get_electron_temperature
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault(
                "current", self.data_e.iloc[:, 1])  # type: ignore
            kwargs.setdefault("V_f", self.results["V_f"]["value"])
            kwargs.setdefault("V_s", self.results["V_s"]["value"])
        elif key == "n_e":
            func = get_electron_density
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault(
                "current", self.data_e.iloc[:, 1])  # type: ignore
            kwargs.setdefault("area", self.probe.area)
            kwargs.setdefault("KT_e", self.results["KT_e"]["value"])
            kwargs.setdefault("I_es", self.results["I_es"]["value"])
            kwargs.setdefault("m_e", self.plasma.m_e)
        elif key == "J_es":
            func = get_electron_saturation_current_density
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault(
                "current", self.data_e.iloc[:, 1])  # type: ignore
            kwargs.setdefault("area", self.probe.area)
            kwargs.setdefault("I_es", self.results["I_es"]["value"])
        elif key == "J_is":
            func = get_ion_saturation_current_density
            kwargs.setdefault("voltage", self.data.iloc[:, 0])  # type: ignore
            kwargs.setdefault("current", self.data.iloc[:, 1])  # type: ignore
            kwargs.setdefault("area", self.probe.area)
            kwargs.setdefault("I_is", self.results["I_is"]["value"])
        elif key == "lambda_De":
            func = get_lambda_D
            kwargs.setdefault("area", self.probe.area)
            kwargs.setdefault("KT_e", self.results["KT_e"]["value"])
            kwargs.setdefault("n", self.results["n_e"]["value"])
        elif key == "sheath":
            func = get_sheath_type
            kwargs.setdefault("lambda_D", self.results["lambda_De"]["value"])
            kwargs.setdefault("radius_probe", self.probe.diameter/2)
        else:
            table = "\n".join(
                [f"{k}\t{v}" for k, v in enumerate(self.results)])
            raise ValueError(
                f"Matching key not found: {key}\nChoose from one of the available options:\n{table}")

        # check for set method
        if method is None:
            method = self.methods[key]

        # solve and return a tuple
        kwargs.setdefault("method", method)
        value, other = func(**kwargs)

        # store values and other returns
        self.results[key]['value'] = value
        self.results[key]['other'] = other

    def find_floating_potential(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'V_f'
        self.find_plasma_property(key, method, **kwargs)

    def find_plasma_potential(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'V_s'
        self.find_plasma_property(key, method, **kwargs)

    def find_ion_saturation_current(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'I_is'
        self.find_plasma_property(key, method, **kwargs)

    def find_ion_density(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'n_i'
        self.find_plasma_property(key, method, **kwargs)

    def find_electron_saturation_current(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'I_es'
        self.find_plasma_property(key, method, **kwargs)

    def find_electron_density(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'n_e'
        self.find_plasma_property(key, method, **kwargs)

    def find_electron_temperature(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'KT_e'
        self.find_plasma_property(key, method, **kwargs)

    def find_electron_saturation_current_density(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'J_es'
        self.find_plasma_property(key, method, **kwargs)

    def find_ion_saturation_current_density(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'J_is'
        self.find_plasma_property(key, method, **kwargs)

    def find_electron_debye_length(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'lambda_De'
        self.find_plasma_property(key, method, **kwargs)

    def find_sheath_type(self, method=None, **kwargs):
        # key for method and saving indexing
        key = 'sheath'
        self.find_plasma_property(key, method, **kwargs)


# Declare available methods
available_algorithms = {
    0: ['topham'],
    1: ['lobbia'],
}
# Declare correspondin functions for available_methods
available_algorithm_functions = {
    0: topham,
    1: lobbia,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
options = method_function_combiner(
    available_algorithms, available_algorithm_functions)



import requests


requests.post('https://api.mynotifier.app', {
    "apiKey": '****-*****-*****',
    "message": "Our first notification!",
    "description": "This is cool",
    "type": "info", # info, error, warning or success
    })





import numpy as np

from pstl.utls.helpers import method_function_combiner, method_selector


def shape_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_shapes, available_shape_functions)
    return func


def cylinderical_ab(*args, **kwargs):
    a = 2*np.sqrt(np.pi)
    b = 0.5
    return a, b


def spherical_ab(*args, **kwargs):
    return _thick_spherical_and_planar_ab()


def planar_ab(*args, **kwargs):
    return _thick_spherical_and_planar_ab()


def _thick_spherical_and_planar_ab(*args, **kwargs):
    a = 1
    b = 1
    return a, b


# Declare available methods
available_shapes = {
    0: ['cylinderial'],
    1: ['spherical'],
    2: ['planar'],
}
# Declare correspondin functions for available_methods
available_shape_functions = {
    0: cylinderical_ab,
    1: spherical_ab,
    2: planar_ab,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
shape_options = method_function_combiner(
    available_shapes, available_shape_functions)



import numpy as np

import pstl.utls.constants as c


def get_electron_density(*args, method=0, **kwargs):
    # Declare available methods
    available_methods = {
        0: 'I_es',
    }

    # Converts method: str -> method: int if method is a str
    if isinstance(method, str):
        reversed_methods = {v: k for k, v in available_methods.items()}
        method = reversed_methods.get(method, None)

    # check for match and get which function to use
    # raises value error with options if failed to match
    if method == 0:  # default
        func = thin_sheath_electron_saturation_current_method
    else:  # makes a table of options if error occurs
        table = "\n".join([f"{k}\t{v}" for k, v in available_methods.items()])
        raise ValueError(
            f"Matching method not found: {method}\nChoose from one of the available options:\n{table}")

    # Call funtion and return result
    return func(*args, **kwargs)


def thin_sheath_electron_saturation_current_method(I_es, area, KT_e, m_e=c.m_e, *args, **kwargs):
    # KT_e [eV]
    inv_root = np.sqrt(2*np.pi*m_e/(KT_e*c.K_B/c.K_B_eV))
    n_e = I_es*inv_root/(area*c.q_e)
    return n_e, {"method": "thin"}



import pyvisa as visa

rm = visa.ResourceManager()

def list_ports():
    rm.list_resources()

def open_port(port):
    resource=rm.open_resource(port)
    return resource

def choose_port():
    
    strin="R"

    # if refresh
    while strin == "R":
        while strin == "R":
            print('\nAvailable Ports:(Enter "R" for refresh)'\
                    +'\nPort #:\tPort Name')
            print('__________________')
            a = 0
            for aa in rm.list_resources():
                a = a + 1
                print('Port ' + str(a) + ':\t' + aa)
            print('__________________')
            strin = input("Enter Port #:\n>>")
        # end of refresh
        try:
            port = rm.list_resources()[int(strin)-1]
        except:
            print("Not a valid entry please try again")
            strin="R"

    print('Trying ' + port)
    resource=open_port(port)
    print('\nFound:')
    print(resource)
    return resource




import numpy as np

from pstl.utls.helpers import method_function_combiner, method_selector


def shape_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_shapes, available_shape_functions)
    return func


def cylinderical_ab(r_p, lambda_D, *args, **kwargs):
    ratio = r_p/lambda_D
    a = 1.18 - 0.00080*np.power(ratio, 1.35)
    b = 0.0684 + np.power(0.722+0.928*ratio, -0.729)
    return a, b


def spherical_ab(r_p, lambda_D, *args, **kwargs):
    ratio = r_p/lambda_D
    a = 1.58 + np.power(-0.056 + 0.816*ratio, -0.744)
    b = -0.933 + np.power(0.0148+0.119*ratio, -0.125)
    return a, b


def planar_ab(r_p, lambda_D, *args, **kwargs):
    ratio = r_p/lambda_D
    a = np.exp(-0.5)*np.sqrt(2*np.pi)*(2.28*np.power(ratio, -0.749))
    b = 0.806*np.power(ratio, -0.0692)
    return a, b


# Declare available methods
available_shapes = {
    0: ['cylinderial'],
    1: ['spherical'],
    2: ['planar'],
}
# Declare correspondin functions for available_methods
available_shape_functions = {
    0: cylinderical_ab,
    1: spherical_ab,
    2: planar_ab,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
shape_options = method_function_combiner(
    available_shapes, available_shape_functions)



import matplotlib.pyplot as plt
import matplotlib.animation as animation
import time
import datetime
import math
import csv
import numpy as np

from pstl.tools.utls.sort import sort_list


class Subplot():
    def __init__(self,ax,*args,**kwargs):
        # assign axes for figure
        self.ax=ax

        # Data list for what to plot are intialized
        self.x=[]
        self.y=[]

        # Start_time for data collection with x-axis in time
        # True: get start time (starts at 0)
        # False: Dont do time as x-axis
        # None: Actually time for each data point
        # Anything else: if in proper datetime format will be start time
        self.start_time=kwargs.get('start_time',True)

        # True if log scale
        self.logx=kwargs.get('logx',False)
        self.logy=kwargs.get('logy',False)

        # plotting kwargs
        plot_kwargs_defaults={
                'marker':'o',
                'linestyle':'-',
                'markerfacecolor':'none'
                }
        self.plot_kwargs=kwargs.get('plot_kwargs',plot_kwargs_defaults)

        # limit_style is either 'magnitude' or None
        # currently limits on x-axis are not yet implemented
        # only ylimit magnitude is implemented
        # magnitude sets how many magnitudes above current magnitude
        # i.e. currentmax  is 4, then magnitude above upper limit would be 10
        self.xlimit_style=kwargs.get('xlimit_style',None)
        self.xlimit=kwargs.get('xlimit',None)
        self.ylimit_style=kwargs.get('ylimit_style',None)
        self.ylimit=kwargs.get('ylimit',None)

        # restrict number of plotted from a list
        self.updatelimit=kwargs.get('updatelimit',None)

        # True to show an update line
        # for updateline_kwargs must be a dictionary
        self.updateline=kwargs.get('updateline',False)
        self.updateline_kwargs=kwargs.get('updateline_kwargs',None)

        # reorder data wrt x data
        self.reorderx=kwargs.get("reorderx",False)

        # function and function_arguments to get updated x,y for
        self.func=kwargs.get('func',None)   # func must return a dict
        self.fargs=kwargs.get('fargs',None) # well need to be a tulpe

        self.location=kwargs.get('location',None) # not sure why here??

        # plot stuff
        self.grid=kwargs.get('grid',True)
        self.xlabel=kwargs.get('xlabel',None)
        self.ylabel=kwargs.get('ylabel',None)
        self.title=kwargs.get('title',None)
        self.rotate_xaxis_labels=kwargs.get("rotate_xaxis_labels",False)

class Figure():
    def __init__(self,nrows:int=1,ncols:int=1,**kwargs):
        # for figure subplot, determins how many rows and cols
        self.nrows=kwargs.get('nrows',nrows)
        self.ncols=kwargs.get('ncols',ncols)

        # if alerts will be used (not yet implemented)
        self.alerts=kwargs.get('alerts',False)

        # Make full screen (true)
        self.full_screen=kwargs.get('full_screen',False)

        # Creates the fig and axes
        fig,ax_list=\
                plt.subplots(nrows=nrows,ncols=ncols)
        self.figure=fig

        if nrows==1 and ncols==1:
            ax=[ax_list]
            nax=1
        else:
            ax=ax_list.ravel()
            nax=len(ax)
        self.axes=ax
        
        # number of axes
        self.nax=nax
        # initialize list of subplots
        subplot=[None]*nax
        for k in range(nax):
            subplot[k]=Subplot(ax[k],**kwargs)
        self.subplot=subplot

    ## This function is called periodically from FuncAnimation
    def _animate(self, i, func=None, fargs=None):

        # This additional function allows data to be collect first
        # and then the function attached to the figure sort the data
        if func is not None and fargs is not None:
            data=func(i,*fargs)
        elif func is not None:
            data=func(i)
        else:
            data=None

        for k in range(self.nax):
            # rename self.subplot for easy
            subplot=self.subplot[k]

            # set xs,ys
            xs=subplot.x
            ys=subplot.y

            # axes to use
            ax=self.axes[k]

            # get data from frunction
            # if,else if fargs need to be passed in
            if data is None:
                if subplot.fargs is not None:
                    out = subplot.func(k,*subplot.fargs)
                else:
                    out=subplot.func(k)
            else:
                if subplot.fargs is not None:
                    out = subplot.func(k,data,*subplot.fargs)
                else:
                    out=subplot.func(k,data)
            x_out=out.get('x_out',None)
            y_out=out.get('y_out',None)
            subplot.fargs=out.get('fargs_out',None)

            # If x-axis is time, then t_out will be used for x_out
            if subplot.start_time is not False:
                if subplot.start_time is True:
                    t_out=datetime.datetime.now()
                    subplot.start_time=t_out
                    t_out=t_out-subplot.start_time
                elif subplot.start_time is None:
                    t_out=datetime.datetime.now()
                else:
                    t_out=datetime.datetime.now()-subplot.start_time

                # round-off museconds
                try:
                    t_out=t_out.strftime('%H:%M:%S.%f')
                except:
                    t_out=str(t_out-datetime.timedelta(
                        microseconds=t_out.microseconds)
                        )
            else: # False (no time dependence)
            # if FAlSE then 
            # no time and t_out will come from either data or index
                t_out=None

            # set t_out to  x_out if x_out is None
            if x_out is None:
                x_out=t_out



            # Add x and y to lists
            xs.append(x_out)
            ys.append(y_out)

            # Restricts x and y lists to last 20 items
            if subplot.updatelimit is not None:
                xs = xs[-int(subplot.updatelimit):]
                ys = ys[-int(subplot.updatelimit):]
            else:
                pass

            # reorder if desired
            if subplot.reorderx:
                xs_plt,ys_plt=sort_list(xs,ys)
            else:
                xs_plt=xs.copy()
                ys_plt=ys.copy()

            # Plot x and y lists (logscale if true)
            ax.clear()
            if subplot.logx and subplot.logy:   # loglog
                ax.loglog(xs_plt, ys_plt,**subplot.plot_kwargs)
            elif subplot.logx:                  # semilogx
                ax.semilogx(xs_plt, ys_plt,**subplot.plot_kwargs)
            elif subplot.logy:                  # semilogy
                ax.semilogy(xs_plt, ys_plt,**subplot.plot_kwargs)
            else:                               # linear plot
                ax.plot(xs_plt, ys_plt,**subplot.plot_kwargs)

            # Plot update line if true
            if subplot.updateline and subplot.updateline_kwargs is not None:
                ax.axvline(xs[-1],**subplot.updateline_kwargs)
            elif subplot.updateline:
                ax.axvline(xs[-1],color='r')
            else:
                pass

            # Update the class x,y with the new xs,ys
            subplot.x=xs
            subplot.y=ys

            # Format plot
            ax.grid(subplot.grid)
            if subplot.rotate_xaxis_labels is not False:
                ax.tick_params(axis='x',labelrotation=subplot.rotate_xaxis_labels)
            #ax.set_xticklabls(ax.get_xticks(),rotation=45)
            ax.set(title=subplot.title)
            ax.set(xlabel=subplot.xlabel)
            ax.set(ylabel=subplot.ylabel)
            if subplot.ylimit_style == 'magnitude' or\
                    subplot.ylimit_style == 'm':
                        if subplot.ylimit is not None:
                            y_max=max(ys);y_min=min(ys)
                            magnitude_max=math.floor(math.log(y_max,10))
                            magnitude_min=math.floor(math.log(y_min,10))
                            #if self.subplot[k].ylimit is list\
                            #        and len(self.subplot[k].ylimit)==2:
                            if len(subplot.ylimit)==2:
                                        ylimit_max=math.pow(10,
                                                magnitude_max+\
                                                subplot.ylimit[1]
                                                )
                                        ylimit_min=math.pow(10,
                                                magnitude_min+\
                                                subplot.ylimit[0]
                                                )
                            elif subplot.ylimit is list\
                                    and len(subplot.ylimit)==1:
                                        ylimit_max=math.pow(10,
                                                magnitude_max+\
                                                subplot.ylimit[0]
                                                )
                                        ylimit_min=math.pow(10,
                                                magnitude_min-\
                                                subplot.ylimit[0])
                            elif subplot.ylimit is int\
                                    or subplot.ylimit is float:
                                        ylimit_max=math.pow(10,
                                                magnitude_max+\
                                                subplot.ylimit
                                                )
                                        ylimit_min=math.pow(10,
                                                magnitude_min-\
                                                subplot.ylimit)
                            else:
                                error_msg="\nERROR: '%s' is not a list, float or int\nit is a %s"%(str(subplot.ylimit),str(type(subplot.ylimit))) 
                                print(error_msg)
                            ax.set(ylim=[ylimit_min,ylimit_max])
                            
    ## Callable function that animates the plot and calls _animate
    def monitor(self,**kwargs):
        """
        kwargs: see matplotlib.animation.FuncAnimation
        defaults built into this class include:
        fig=self.figure (from this Figure class)
        func=self._animate (from this Figure class)
            THIS IS NOT the updating data function
            That is seperatly called inside the self._animate
            func from the Subplot class
        fargs=None as nothing else is passed to self._animate
            as the self._animate function calls its own function
            and fargs for data collection from the Subplot class
            To pass in the func and fargs to _animate, set:
            fargs=(yourFunc,(yourArgs)) a tulpe inside a tulpe
        """
        # set defaults in kwargs
        fig=kwargs.pop('fig',self.figure)
        func=kwargs.pop('func',self._animate)
        kwargs['fargs']=kwargs.get('fargs',None)
        # Set up plot to call animate() function periodically
        ani = animation.FuncAnimation(fig,func,**kwargs)
        # to do full screen
        wm = plt.get_current_fig_manager()
        if self.full_screen:
            wm.window.state('zoomed')
        # update & show plot
        plt.show()


    def save_all(self,**kwargs):
        for k in range(self.nax):
            x=self.subplot[k].x
            y=self.subplot[k].y
            xlabel=self.subplot[k].xlabel
            ylabel=self.subplot[k].ylabel
            csv_out=np.vstack((x,y)).transpose()
            header=xlabel+","+ylabel
            fname=kwargs.pop('fname','monitor_subplt_%i_data.csv'%(k))
            kwargs['header']=kwargs.get('header',header)
            kwargs['delimiter']=kwargs.get('delimiter',',')

            print("saving '%s'..."%(str(fname)))

            np.savetxt(fname,csv_out,**kwargs)



import numpy as np

# some functions are dependent on classes defined elsewhere
# here they are passed in and functions are called

def setupDAQProbeChannels():
    channels=[1,2,9,10,11,12]
    resistance=[1,1,1,1,1]
    return channels,resistance

def scanDAQ(daq,slot,channels,func=None):
    if func is None:
        func=daq.get
    output=np.zeros_like(channels,dtype=float)
    for j in range(len(channels)):
        channel=channels[j]
        # splice is to ignore '\n'
        # then convert to float
        location=int(np.add(np.multiply(slot,100),channel))
        output[j]=func(location)[:-1]
    return output

def scanDAQRESVDC(daq,slot,channel):
    r=scanDAQ(daq,slot,channel,daq.getRes)
    vr=scanDAQ(daq,slot,channel,daq.getVDC)
    return r,vr

def scanDAQVDC(daq,slot,channel):
    return scanDAQ(daq,slot,channel,daq.getVDC)



from pstl.instruments.daq.agilent import commands as cmds


class CHANNEL():
    def __init__(self,location,chtype="VDC"):
        self.location=location
        self.change(location,chtype)

        
    def change(self,location,chtype):
        """
        Changes channel type
        """
        if chtype=="VDC":
            get=cmds.cmdGetVoltageDC(self.location)
        elif chtype=="VAC":
            get=cmds.cmdGetVoltageAC(self.location)
        elif chtype=="TCK":
            get=cmds.cmdGetTemperatureTCK(self.location)
        elif chtype=="RES":
            get=cmds.cmdGetResistance(self.location)
        else:
            print("Channel Type (chtype) not set to known value at Channel %s"%(str(self.location)))
            get=None

        self.chtype=chtype
        self.getcmd=get



import numpy as np

from pstl.utls.verify import verify_1D_array, verify_type, verify_polarity


def verify_key_exist(key, dictionary):
    if key in dictionary:
        return True
    else:
        return False


def set_savgol_filter_defaults(ydata: np.ndarray,
                               window_length: int | None = None,
                               polyorder: int | None = None,
                               savgol_filter_kwargs: dict = {},
                               **kwargs):
    # Verify 1D array and get length of ydata
    verify_1D_array(ydata)
    len_ydata = len(ydata)

    if window_length is None:
        default_window_length = int(0.08*len_ydata)
        default_window_length = default_window_length if default_window_length % 2 != 0 else default_window_length+1
        window_length = savgol_filter_kwargs.pop(
            'window_length', default_window_length)
    else:
        _ = savgol_filter_kwargs.pop('window_length', None)

    if polyorder is None:
        polyorder = savgol_filter_kwargs.pop('polyorder', 1)
    else:
        _ = savgol_filter_kwargs.pop('polyorder', None)

    # verify int and verify if positive
    verify_type(window_length, int, 'window_length')
    verify_type(polyorder, int, 'polyorder')
    verify_polarity(window_length)
    verify_polarity(polyorder)

    return window_length, polyorder, savgol_filter_kwargs



import smtplib
from email.message import EmailMessage
import requests

def send(to,subject, body, Sender):
    """
    Sender is a class with subclasses EMAIL and API
    attributes=example
    Sender.Email.address='sample@gmail.com'
    Sendr.Email.password='third-party app password'
    Sender.Email.smtp='smtp.gmail.com'
    Sender.Email.port=587
    """

    msg=EmailMessage()
    msg.set_content(body)
    msg['subject']=subject
    msg['to']=to

    msg['from']=Sender.Email.address
    password=Sender.Email.password
    server = smtplib.SMTP(Sender.Email.smtp,Sender.Email.port)

    server.starttls()
    server.login(Sender.Email.address, Sender.Email.password)
    server.send_message(msg)

    server.quit()

def text(to,subject, body, Sender):
    send(to,subject,body,Sender)

def email(to,subject, body, Sender):
    send(to,subject,body,Sender)

def api(to,subject,body,Sender):
    """
    Sender is a class with subclasses EMAIL and API
    attributes=example
    Sender.Api.app='https://api.mynotifier.app'
    Sender.Api.msgtype="info"

    """
    requests.post(Sender.Api.app, {
        "apiKey": to,
        "message": subject,
        "description": body,
        "type": Sender.Api.msgtype, # info, error, warning or success
        })




import numpy as np

from pstl.utls import constants as c
from .utls.ion_current import thin, thick, transitional
from pstl.utls.helpers import method_selector
from pstl.utls.helpers import method_function_combiner
default_fit_thin_kwargs = {
    'deg': 1, 'power': 1, 'polarity': 1,
    'reverse': False, 'return_best': True, 'fit_type': "linear",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': None, 'fstep_type': None, 'fstep_adjust': True, 'fitmax': None,
    'bstep': None, 'bstep_type': None, 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.30,
    'threshold_rsq': 0.95, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 1.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}


def find_ion_current(shape, *args, method=None, **kwargs):
    value, _ = get_ion_current(shape, method=method, *args, **kwargs)
    return value


def get_ion_current(shape, *args, method=None, ** kwargs):

    # set default -> (thin)
    method = 0 if method is None else method

    # select function based on integer_method
    func = method_selector(method, available_methods, available_functions)

    return func(shape, *args, **kwargs)


def check_for_ion_current_fit(I_i_fit, voltage, current, shape=None, method=None, **kwargs):
    # determine starting point (all positive after vf) if not given
    if I_i_fit is None:
        # determine starting point (all positive after V_f)
        fit_kwargs = kwargs.pop('fit_kwargs', {})
        # get floating potential
        I_i, extras = get_ion_current(
            shape, voltage, current, method=method, fit_kwargs=fit_kwargs,
        )
        I_i_fit = extras["fit"]

    return I_i_fit


def thin_method(shape, *args, **kwargs):
    func = thin.shape_fit_selector(shape)
    return func(*args, **kwargs)


def thin_func_method(shape, *args, **kwargs):
    func = thin.shape_func_selector(shape)
    return func(*args, **kwargs)


def transitional_method(shape, *args, **kwargs):
    func = transitional.shape_selector(shape)
    return func(*args, **kwargs)


def thick_method(shape, *args, **kwargs):
    func = thick.shape_selector(shape)
    return func(*args, **kwargs)


# Declare available methods
available_methods = {
    0: ['thin'],
    1: ['transitional'],
    2: ['thick', 'OML'],
    3: ['thin-func']
}
# Declare correspondin functions for available_methods
available_functions = {
    0: thin_method,
    1: transitional_method,
    2: thick_method,
    3: thin_func_method,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
options = method_function_combiner(available_methods, available_functions)



import time
import json

import numpy as np

from pstl.instruments.daq.agilent.gpib import Agilent34970A as DAQ
from pstl.instruments.ps.kepco.gpib import BOP_100_1M_488B as PS
from pstl.tools.animate.monitor import Figure as FIG 
from pstl.scripts.janus.utls import sweep
from pstl.scripts.janus.utls.agilent import scanDAQRESVDC

def exit_handler(save,obj):
    if save:
        obj.save_all()

def get_parameters(fname=None):
    if fname is not None:
        with open(fname,'r') as f:
            params=jsons.load(fname)

    else:
        voltage_delay=1
        daq_slot=1
        ps_channel=[1]
        r_channel=[2,9,10,11,12]
        channels=ps_channel+r_channel
        resistance=[1,1,1,1,1]
        #supplyVoltages
        vstart=-5
        vstop=5
        dv=1
        supply_voltages=np.arange(vstart,vstop+dv,dv)
        params={
                'voltage_delay':voltage_delay,
                'daq_slot':daq_slot,
                'ps_channel':ps_channel,
                'r_channels':r_channel,
                'channels':channels,
                'resistance':resistance,
                'supply_voltages':supply_voltages
                }
    return params



def main():
    # get parameters 
    parameters_fname=None   # if not None, read in this json file
    params=get_parameters(parameters_fname)

    # retrive needed variables from parameters and set standard defaults
    voltage_delay=params.get('voltage_delay',1) # seconds
    daq_slot=params.get('daq_slot',1) 
    channels=params.get('channels',None)
    resistance=params.get('resistance',True) # True: will probe
    supply_voltages=params.get('supply_voltages',None)
    nrows=params.get("nrows",5)
    ncols=params.get("ncols",1)
    save=params.get("save",True)

    # check required inputs and raise error if not provided
    if channels is None: 
        raise ValueError("'channels' parameter was not defined")
    if supply_voltages is None: 
        raise ValueError("'supply_voltages' parameter was not defined")
    

    # setup DAQ class and add card class
    daq=DAQ("GPIB0::10::INSTR")
    daq.addCardAgilent34901A(daq_slot,20,"VDC")
    # setup power supply class
    ps=PS("GPIB0::6::INSTR")

    # setup figure for data
    kwargs={
            "func":sweep.update,
            "xlabel":"V [V]",
            "ylabel":"I [A]",
            "updatelimit":len(supply_voltages),
            "reorderx":True,
            "updateline":True,
            "full_screen":True
            }
    langmuir=FIG(nrows,ncols,**kwargs)
    """
    voltages    Channels
    -1          2       9   10  etc.
    0
    -1

    """

    # get resistance before stating loop
    if resistance is True:
        resistance=scanDAQRESVDC(daq,daq_slot,channels)

    # run actual program
    print("running ...")
    print("close figure to stop and save data")
    try:
        # args for scan function in animate
        args=(ps,daq,daq_slot,channels,supply_voltages,resistance,voltage_delay)
        # pass dict to monitor for animate function
        kwargs={
                "fargs":(sweep.scan,args)
                }
        langmuir.monitor(**kwargs)
    except KeyboardInterrupt:
        print("exiting ...")
    exit_handler(save,langmuir)


if __name__=="__main__":
    main()



import numpy as np

from pstl.diagnostics.probes.langmuir.single.analysis.utls.ion_current.thin import _function
from pstl.utls import constants as c
from pstl.utls.decorators import where_function_else_zero
from pstl.utls.helpers import method_function_combiner
from pstl.utls.helpers import method_selector as method_selector
from pstl.utls.helpers import make_CustomFit


def shape_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_methods, available_functions)
    return func


def cylinderical_method(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_D, *args, **kwargs):

    func = where_function_else_zero(
        _cylinderical_function, cylinderical_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i, r_p, lambda_D)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "transitional", "shape": "cylinderical", "fit": fit}
    return I_i, extras


def spherical_method(voltage, area, n0, V_s, m_i, KT_e,  r_p, lambda_D, *args, **kwargs):
    func = where_function_else_zero(
        _spherical_function, spherical_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i, r_p, lambda_D)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "transitional", "shape": "spherical", "fit": fit}
    return I_i, extras


def planar_method(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_D, *args, **kwargs):
    func = where_function_else_zero(
        _planar_function, planar_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i, r_p, lambda_D)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "transitional", "shape": "planar", "fit": fit}
    return I_i, extras


# domain conditions
def cylinderical_domain_condition(voltage, V_s, KT_e):
    return _cylinderical_spherical_domain_condition(voltage, V_s, KT_e)


def spherical_domain_condition(voltage, V_s, KT_e):
    return _cylinderical_spherical_domain_condition(voltage, V_s, KT_e)


def planar_domain_condition(voltage, V_s, KT_e):
    condition = np.divide(
        np.subtract(V_s, voltage),  # [V]
        KT_e                         # [eV]
    )

    returns = np.where

    cond1 = 3 < condition
    cond2 = condition < 30
    return np.logical_and(cond1, cond2)  # 10 < rp/l < 45

# generalized domain


def _cylinderical_spherical_domain_condition(voltage, V_s, KT_e):
    condition = np.divide(
        np.subtract(V_s, voltage),  # [V]
        KT_e                         # [eV]
    )
    return condition > 1


# Functions for calculating current
def cylinderical_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_D):
    ratio = r_p/lambda_D
    a = 1.18 - 0.00080*np.power(ratio, 1.35)
    b = 0.0684 + np.power(0.722+0.928*ratio, -0.729)
    return _partial_function(voltage, area, n0, V_s, m_i, KT_e, a, b)


def spherical_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_D):
    ratio = r_p/lambda_D
    a = 1.58 + np.power(-0.056 + 0.816*ratio, -0.744)
    b = -0.933 + np.power(0.0148+0.119*ratio, -0.125)
    return _partial_function(voltage, area, n0, V_s, m_i, KT_e, a, b)


def planar_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_D):
    ratio = r_p/lambda_D
    a = np.exp(-0.5)*np.sqrt(2*np.pi)*(2.28*np.power(ratio, -0.749))
    b = 0.806*np.power(ratio, -0.0692)
    partial = _partial_function(voltage, area, n0, V_s, m_i, KT_e, a, b)
    standard = _function(area, n0, m_i, KT_e)
    return np.add(partial, standard)


# Functions for calculating current (position arguments important for wrappers here) #######
def _cylinderical_function(voltage, V_s, KT_e, area, n0, m_i, r_p, lambda_D):
    return cylinderical_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_D)


def _spherical_function(voltage, V_s, KT_e, area, n0, m_i, r_p, lambda_D):
    return spherical_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_D)


def _planar_function(voltage, V_s, KT_e, area, n0, m_i, r_p, lambda_D):
    return planar_function(voltage, area, n0, V_s, m_i, KT_e, r_p, lambda_D)
#############################################################################################


# generalized
def _partial_function(voltage, area, n0, V_s, m_i, KT_e, a, b):
    A = (c.e*n0*area)*np.sqrt(c.e*KT_e/(2*np.pi*m_i))*a

    x = np.divide(V_s-voltage, KT_e)

    with np.errstate(invalid="ignore"):
        power_x = np.power(x, b)

    current = np.multiply(
        A,
        power_x
    )

    return current


# Declare available methods
available_methods = {
    0: ['cylinderical'],
    1: ['spherical'],
    2: ['planar', 'planer'],
}
# Declare correspondin functions for available_methods
available_functions = {
    0: cylinderical_method,
    1: spherical_method,
    2: planar_method,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
options = method_function_combiner(available_methods, available_functions)



from pstl.tools.alerts import notification

class MESSAGE():
    def __init__(self,/,subject=None,body=None,msg_type="info"):
        self.subject=subject
        self.body=body
        self.msg_type=msg_type

    def send(self,to:str,sender):




from pstl.instruments.daq import agilent

def open_port(port=None):

    instrument=agilent.gpib.Agilent34970A(port)

    return instrument

def main():
    port="GPIB0::10::INSTR"
    card=1
    channel=13
    location=int(card*100+channel)
    port=None
    daq=open_port(port)
    daq.addCardAgilent34901A(1,20,'TCK')
    daq.list_cards()
    #daq.card[1].list_channels()
    print(daq.card[card].channel[channel].getcmd)
    r=daq.get(location)
    print("Temperature %.2f degC"%(float(r)))
    r=daq.get(card,channel)
    print("Temperature %.2f degC"%(float(r)))

if __name__=="__main__":
    main()



import numpy as np

from pstl.utls import constants as c
from pstl.utls.helpers import method_function_combiner
from pstl.utls.helpers import method_selector as method_selector
from pstl.utls.helpers import FunctionFit, CustomFit, make_CustomFit
from pstl.utls.decorators import where_function_else_zero


def shape_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_methods, available_functions)
    return func


def cylinderical_method(voltage, area, n0, V_s, m_i, KT_e, *args, **kwargs):

    func = where_function_else_zero(
        _cylinderical_function, cylinderical_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "thick", "shape": "cylinderical", "fit": fit}
    return I_i, extras


def spherical_method(voltage, area, n0, V_s, m_i, KT_e, *args, **kwargs):
    func = where_function_else_zero(
        _spherical_function, spherical_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "thick", "shape": "spherical", "fit": fit}
    return I_i, extras


def planar_method(voltage, area, n0, V_s, m_i, KT_e, *args, **kwargs):

    func = where_function_else_zero(
        _planar_function, planar_domain_condition)

    coefs = (V_s, KT_e, area, n0, m_i)
    I_i = func(voltage, *coefs)

    fit = make_CustomFit(func, voltage, I_i, coefs)

    # make returns
    extras = {"method": "thick", "shape": "planar", "fit": fit}
    return I_i, extras


# domain conditions
def cylinderical_domain_condition(voltage, V_s, KT_e):
    return _domain_condition(voltage, V_s, KT_e)


def spherical_domain_condition(voltage, V_s, KT_e):
    return _domain_condition(voltage, V_s, KT_e)


def planar_domain_condition(voltage, V_s, KT_e):
    return _domain_condition(voltage, V_s, KT_e)

# generalized domain


def _domain_condition(voltage, V_s, KT_e):
    condition = np.divide(
        np.subtract(V_s, voltage),  # [V]
        KT_e                         # [eV]
    )
    return condition > 1  # >>1


# Functions for calculating current (position arguments important for wrappers here) #######
# make these more robust by changing position arguments to keyword arguments via dictionary
def _cylinderical_function(voltage, V_s, KT_e, area, n0, m_i):
    return cylinderical_function(voltage, area, n0, V_s, m_i)


def _spherical_function(voltage, V_s, KT_e, area, n0, m_i):
    return spherical_function(voltage, area, n0, V_s, m_i, KT_e)


def _planar_function(voltage, V_s, KT_e, area, n0, m_i):
    return planar_function(voltage, area, n0, V_s, m_i, KT_e)
#############################################################################################


def cylinderical_function(voltage, area, n0, V_s, m_i):
    A = (c.e*n0*area/np.pi)*np.sqrt(2*c.e/m_i)

    with np.errstate(invalid="ignore"):
        sqrt_x = np.sqrt(V_s-voltage)

    current = np.multiply(
        A,
        sqrt_x
    )

    return current


def spherical_function(voltage, area, n0, V_s, m_i, KT_e):
    return _spherical_and_planar_function(voltage, area, n0, V_s, m_i, KT_e)


def planar_function(voltage, area, n0, V_s, m_i, KT_e):
    return _spherical_and_planar_function(voltage, area, n0, V_s, m_i, KT_e)

# generalized


def _spherical_and_planar_function(voltage, area, n0, V_s, m_i, KT_e):
    A = (c.e*n0*area)*np.sqrt(c.e*KT_e/(2*np.pi*m_i))

    current = np.multiply(A, np.divide(V_s-voltage, KT_e))

    return current


# Declare available methods
available_methods = {
    0: ['cylinderical'],
    1: ['spherical'],
    2: ['planar', 'planer'],
}
# Declare correspondin functions for available_methods
available_functions = {
    0: cylinderical_method,
    1: spherical_method,
    2: planar_method,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
options = method_function_combiner(available_methods, available_functions)



import os.path
import json

from pstl import contacts
from pstl.contacts import individuals
from pstl.contacts import groups
from pstl.tools.alerts import send

cdir=os.path.dirname(contacts.__file__)
idir=os.path.dirname(individuals.__file__)
gdir=os.path.dirname(groups.__file__)

def get_json(filename,suffix=".json",wdir=cdir):
    try: 
        filepath=os.path.join(wdir,filename+suffix)
        with open(filepath) as f:
            data=json.load(f)
    except TypeError:
        print("%s was not inputted as a string"%(str(last)))
    return data


def get_individual(name,/,suffix=".json",wdir=idir):
    return get_json(name,suffix,wdir)

def get_group(name,/,suffix=".json",wdir=gdir):
    return get_json(name,suffix,wdir)

def test_if_known(i,what):
    if i is None or i == "" or i == "unknown"\
            or i == "Unknown" or i == "UNKNOWN":
                print("%s type is not known"%(what))
                return False
    else:
        return True

def method_all():
    return ['text','email','api']
    

def individual(name,method,/,suffix=".json",wdir=None):
    if wdir is None: wdir=idir
    if type(method) is list: pass
    else: method = [method]

    info=get_individual(name,suffix,wdir=wdir)
    try:
        #VIA=[]
        VIA={}
        if "all" in method or "All" in method or "ALL" in method:
            method = method_all()
        if "text" in method:
            if 'phone' in info:
                ip=info['phone']
                if 'number' in ip and 'service' in ip:
                    ips=ip['service']
                    if test_if_known(ips,'Service'):
                        # service data
                        sdata=get_json("sms")
                        if ips in sdata:
                            text_extention=sdata[ips]
                            #VIA.append({'text':ip['number']+text_extention})
                            VIA['text']=ip['number']+text_extention
                        else:
                            print("please update your sms.json file")
        if "email" in method:
            if 'email' in info:
                ie=info['email']
                if test_if_known(ie,'Email'):
                    #VIA.append({'email':ie})
                    VIA['email']=ie

        if "api" in method:
            if 'api' in info:
                ia=info['api']
                if test_if_known(ia,'Api'):
                    #VIA.append({'api':ia})
                    VIA['api']=ia
    except:
        print("error: '%s' could not be performed"%(str(method)))
        print("try: 'text', 'email' or 'api'")

    return {name:VIA}

def group(groupname,method,/,\
        groupsuffix=".json",individualsuffix=".json",\
        wdir=None,widir=None):
    if wdir is None: 
        wdir=gdir
    if widir is None:
        widir=idir
    if type(method) is list: pass
    else: method = [method]

    groupinfo=get_group(groupname,suffix=groupsuffix,wdir=wdir)
    PEOPLE={}
    try:
        if "all" in method or "All" in method or "ALL" in method:
            method = method_all()
        for name in groupinfo['people']:
            METHOD=[]
            for each in method:
                if each in groupinfo['people'][name]:
                    METHOD.append(each)
            person=individual(name,METHOD,suffix=individualsuffix,wdir=widir)
            PEOPLE[name]=person[name]
    except:
        print("Error: in contactGroup")

    return PEOPLE

def contact_individual(Sender,msg,name,method,/,subject=None,suffix=".json",wdir=None):
    """
    User is a class. either User class for text or email,
    or Api class for api
    """
    if wdir is None: wdir=idir
    if type(method) is list: pass
    else: method = [method]

    info=individual(name,method,suffix=suffix,wdir=wdir)
    for name in info:
        for via in info[name]:
            if via == 'text':
                # assign text function here
                func=send.text
            if via == 'email':
                # assign email function here
                func=send.email
            if via == 'api':
                # assign api function here
                func=send.api

            address=info[name][via]
            # call via function
            func(address,subject,msg,Sender)
            print(address)


def contact_group(Sender,msg,groupname,method,/,subject=None,\
        groupsuffix=".json",individualsuffix=".json",wdir=None,widir=None):
    if wdir is None:
        wdir=gdir
    if widir is None:
        widir=idir
    if type(method) is list: pass
    else: method = [method]

    groupinfo=group(groupname,method,\
            groupsuffix=groupsuffix,individualsuffix=individualsuffix,\
            wdir=wdir)
    for name in groupinfo:
        for via in groupinfo[name]:
            if via == 'text':
                # assign text function here
                func=send.text
            if via == 'email':
                # assign email function here
                func=send.email
            if via == 'api':
                # assign api function here
                func=send.api

            address=groupinfo[name][via]
            # call via function
            func(address,subject,msg,Sender)

def email_sender(name,/,suffix=".json",wdir=idir):
    info=get_json(name,suffix=suffix,wdir=wdir)
    if 'email' in info:
        user_email=info['email']
        if 'password' in info:
            if 'app' in info['password']:
                user_password=info['password']['app']
        if 'smtp' in info:
            user_smtp=info['smtp']
    return EMAIL(user_email,user_password,user_smtp)

def api_sender(app,msgtype,/,app_site=False,suffix=".json",wdir=idir):
    if app_site:
        pass
    else:
        info=get_json(app,suffix=suffix,wdir=wdir)
        if 'app' in info:
            app_site=info['app']
    return API(app_site,msgtype)

def sender(name=None,app=None,msgtype='info',/,suffix=".json",wdir=idir):
    if name is not None:
        Email=email_sender(name,suffix=suffix,wdir=wdir)
    else:
        Email=EMAIL(None,None,None,None,None)
    if app is not None:
        Api=api_sender(app,msgtype)
    else:
        Api=API(None,None)
    return SENDER(Email=Email,Api=Api)


class EMAIL():
    def __init__(self,address,password,smtp,port=587):
        self.address=address
        self.password=password
        self.smtp=smtp
        self.port=port

class API():
    def __init__(self,app,msgtype):
        self.app=app
        self.msgtype=msgtype

class SENDER():
    def __init__(self,/,Email=None,Api=None,\
            email=None,password=None,smtp=None,port=587,\
            app=None,msgtype=None):
        if Email is None:
            self.Email=EMAIL(email,password,smtp,port)
        else:
            self.Email=Email
        if Api is None:
            self.Api=API(app,msgtype)
        else:
            self.Api=Api



from pstl.utls.protocol.gpib.pyvisa import Open
from pstl.instruments.ps.kepco import commands as cmds

class BOP(Open):
    """
    class for kepco bipolar operational DC power supply using ieee 488B gpib

    setCurrent() function has not been tested
    """
    def __init__(self,port=None,**kwargs):

        Open.__init__(self,port)

        self.ieee=kwargs.get("ieee","")
        self.class_name=kwargs.get('class_name','kepco_BOP')
        self.name=kwargs.get('name',self.class_name)
        self.type="ps"
        self.vmax=kwargs.get('vmax',None)
        if self.vmax is None: 
            self.vmax=float(input("Please enter max voltage [V]"))
        self.imax=kwargs.get('imax',None)
        if self.imax is None:
            self.imax=float(input("Please enter max current [A]"))

class BOP_100_1M(BOP):
    """
    Builds off the BOP class adds vmax=100 [V] and imax=1 [A]
    """
    def __init__(self,port=None,**kwargs):

        kwargs['class_name']=kwargs.get('class_name','kepco_BOP_100_1M')
        kwargs['name']=kwargs.get('name',kwargs['class_name'])
        kwargs['vmax']=100;kwargs['imax']=1
        BOP.__init__(self,port,**kwargs)

class BOP_100_1M_488B(BOP_100_1M):
    """
    Builds off the BOP_100_1M class adds ieee=488B
    """
    def __init__(self,port=None,**kwargs):

        kwargs['class_name']=kwargs.get('class_name','kepco_BOP_100_1M_488B')
        kwargs['name']=kwargs.get('name',kwargs['class_name'])
        kwargs['ieee']='488B'
        BOP_100_1M.__init__(self,port,**kwargs)

        self.vhexalen=kwargs.get('vhexalen',3)
        self.ihexalen=kwargs.get('ihexalen',2)
        self.vmaxhexa=self.getHex(self.vmax,self.vmax,self.vhexalen)
        self.imaxhexa=self.getHex(self.imax,self.imax,self.ihexalen)
        

    def getHex(self,num,maximum,hexalen):
        return cmds.convert2hex(num,maximum,hexalen)

    def getCMDVoltage(self,voltage,polarity,**kwargs):

        vmax=kwargs.get("vmax",self.vmax)
        vhexalen=kwargs.get("vhexalen",self.vhexalen)
        imax=kwargs.get("imax",self.imax)
        ihexalen=kwargs.get("ihexalen",self.ihexalen)
        
        LOWER=False
        if voltage <= .1*vmax:
            vmax=vmax/10
            LOWER=True

        hexaVoltage=self.getHex(voltage,vmax,vhexalen)
        hexaCurrent=self.getHex(imax,imax,ihexalen)
        polarity=cmds.polarity(polarity)
        if polarity==1:
            if LOWER:
                hexaPolarity="2"
            else:
                hexaPolarity="0"
        elif polarity==-1:
            if LOWER:
                hexaPolarity="3"
            else:
                hexaPolarity="1"
        else:
            raise NotImplementedError("Unknown %s"%(str(polarity)))
        cmd=hexaPolarity+hexaVoltage+hexaCurrent
        return cmd

    def setVoltage(self,voltage,**kwargs):
        polarity=kwargs.get('polarity',None)
        if polarity is None: 
            if voltage>=0: 
                polarity=1 
            else: 
                polarity=-1
                voltage=abs(voltage)
        self.write(self.getCMDVoltage(float(voltage),polarity,**kwargs))

    def getCMDCurrent(self,current,polarity,**kwargs):

        vmax=kwargs.get("vmax",self.vmax)
        vhexalen=kwargs.get("vhexalen",self.vhexalen)
        imax=kwargs.get("imax",self.imax)
        ihexalen=kwargs.get("ihexalen",self.ihexalen)
        
        LOWER=False
        if current <= .1*imax:
            imax=imax/10
            LOWER=True

        hexaVoltage=self.getHex(vmax,vmax,vhexalen)
        hexaCurrent=self.getHex(current,imax,ihexalen)
        polarity=cmds.polarity(polarity)
        if polarity==1:
            if LOWER:
                hexaPolarity="6"
            else:
                hexaPolarity="4"
        elif polarity==-1:
            if LOWER:
                hexaPolarity="7"
            else:
                hexaPolarity="5"
        else:
            raise NotImplementedError("Unknown %s"%(str(polarity)))
        cmd=hexaPolarity+hexaVoltage+hexaCurrent
        return cmd

    def setCurrent(self,current,**kwargs):
        polarity=kwargs.get('polarity',None)
        if polarity is None: 
            if current>=0: 
                polarity=1 
            else: 
                polarity=-1
                current=abs(current)
        self.write(self.getCMDCurrent(float(current),polarity,**kwargs))



from typing import Optional, Union, Tuple, Any, List

import numpy as np
import numpy.typing as npt


def verify_iterables(iter1: npt.ArrayLike, iter2: npt.ArrayLike) -> None:
    iter1 = np.array(iter1)
    iter2 = np.array(iter2)
    arr1_shape = np.shape(iter1) if hasattr(
        iter1, "__len__") else np.shape(np.array(iter1))
    arr2_shape = np.shape(iter2) if hasattr(
        iter2, "__len__") else np.shape(np.array(iter2))

    if arr1_shape != arr2_shape:
        raise ValueError('Iterables must have the same shape')
    elif len(iter1) != len(iter2):
        raise ValueError('Iterables must have the same length')


def verify_polarity(x: Union[int, float], polarity: str = "positive") -> None:
    if polarity == "positive":
        if x >= 0:
            pass
        else:
            raise ValueError(
                "Argument is negative but should be positve: %s" % (str(x)))
    if polarity == "negative":
        if x <= 0:
            pass
        else:
            raise ValueError(
                "Argument is positive but should be negative: %s" % (str(x)))


def verify_array_polarity(xdata: npt.ArrayLike, polarity: Union[str, List[str]] = "positive") -> None:
    xdata = np.array(xdata)
    if isinstance(polarity, str):
        polarity = [polarity]*len(xdata)
    verify_type(polarity, list)
    for k in range(0, len(xdata), 1):
        val = xdata[k]
        pol = polarity[k]
        verify_polarity(val, polarity=pol)


def verify_1D_array(array: npt.ArrayLike) -> None:
    np_arr = np.array(array)
    if np_arr.ndim != 1:
        raise ValueError("array is not 1D: %sD" % (str(np_arr.ndim)))


def verify_type(x: npt.ArrayLike, var_type: Any, var_str: Optional[str] = None, none_type: bool = False) -> None:
    """
    Verifys that the input 'x' is of type 'var_type'. Raises error if not a match. 
    If 'var_str' is defined, the print function will say:

    '<var_str>' is not type '<var_type>': <x>

    Parameters
    ----------
    x : any
        Variable to test if matches 'var_type'
    var_type : variable type, or tuple of types (cannot be None)
        Variable type to see if matches 'x'
    var_str : str, optional
        String of variable name to dispaly if Raise TypeError occurs
    none_type :  bool, optional
        If True, will also allow variable type to be None

    Returns
    -------
    None

    Rasies
    ------
    TypeError
        If 'x' does not match type of 'var_type'
    """
    if var_str is None:
        if isinstance(x, var_type) is False:
            if none_type and x is None:
                pass
            elif none_type:
                raise TypeError("Argument is not %s or None: %s" %
                                (str(var_type), str(type(x))+": " + str(x)))
            else:
                raise TypeError("Argument is not %s: %s" %
                                (str(var_type), str(type(x))+": " + str(x)))
    elif isinstance(var_str, str):
        if isinstance(x, var_type) is False:
            if none_type and x is None:
                pass
            elif none_type:
                raise TypeError("'%s' is not %s or None: %s" %
                                (var_str, str(var_type), str(type(x))+": " + str(x)))
            else:
                raise TypeError("'%s' is not %s: %s" %
                                (var_str, str(var_type), str(type(x))+": " + str(x)))
    elif not isinstance(var_str, str):
        raise TypeError("'var_str' is not %s: %s" %
                        (str(str), str(x)))
    else:
        raise Exception("Something went wrong in verify_type()")


def verify_pair_of_1D_arrays_type_and_shape(
        # Union[npt.NDArray, Sequence[Union[int, float]]],
        xdata: npt.ArrayLike,
        ydata: npt.ArrayLike  # Union[npt.NDArray, Sequence[Union[int, float]]]
    # ) -> Tuple[Union[npt.NDArray, Sequence[Union[int, float]]], Union[npt.NDArray, Sequence[Union[int, float]]]]:
) -> Tuple[npt.ArrayLike, npt.ArrayLike]:
    """
    Verifys that the two arrays are equal 1D shape and correct orientation (#,). Raises ValueError if not.

    Parameters
    ----------
    xdata : arraylike
        experimental data to check if shape is equilvelent to ydata
    ydata : arraylike
        experimental data to check if shape is  equilvelent to xdata

    Returns
    ----------
    (xdata : np.ndarray, ydata : np.ndarray)
        If input arguments are not numpy arrays, they will be converted to np.ndarray. Otherwise, the original arrays are returned.


    Raises
    ------
    ValueError
        If xdata and ydata are not the same shape
    """
    # check if xdata and ydata are np.ndarray, if not convert them to
    if isinstance(xdata, np.ndarray) is False:
        try:
            xdata = np.array(xdata, dtype=np.float64)
            if xdata.ndim != 1:
                raise TypeError("'xdata' is not 1D array")
        except ValueError as errmsg:
            raise ValueError(str(errmsg)+" in 'xdata'")
    if isinstance(ydata, np.ndarray) is False:
        try:
            ydata = np.array(ydata, dtype=np.float64)
            if ydata.ndim != 1:
                raise TypeError("'ydata' is not 1D array")
        except ValueError as errmsg:
            raise ValueError(str(errmsg)+" in 'ydata'")
    return xdata, ydata


def verify_pair_of_1D_arrays_lengths(xdata: npt.ArrayLike, ydata: npt.ArrayLike):
    """
    Verifys that the two arrays are equal length. Raises ValueError if not.

    Parameters
    ----------
    xdata : arraylike
        experimental data to check if lengths are equilvelent to ydata
    ydata : arraylike
        experimental data to chekc if length is  equilvelent to xdata

    Returns
    ----------
    None

    Raises
    ------
    ValueError
        If xdata and ydata are not the same length
    """
    # convert to numpy arrays s.t. that lengths can be determined
    ydata = np.array(ydata)
    xdata = np.array(xdata)
    # get lengths of data and verify they are the same lengths.
    len_ydata = len(ydata)
    len_xdata = len(xdata)
    if len_xdata != len_ydata:
        raise ValueError("xdata and ydata arrays must have the same length.")


def verify_pair_of_1D_arrays(xdata: npt.ArrayLike, ydata: npt.ArrayLike) -> Tuple[npt.ArrayLike, npt.ArrayLike]:
    """
    Verifys that the two arrays are equal 1D shape (length) and correct orientation (#,). 
    Raises ValueError if not.

    Parameters
    ----------
    xdata : arraylike
        experimental data to check if equilvelent to ydata
    ydata : arraylike
        experimental data to check if equilvelent to xdata

    Returns
    ----------
    (xdata : np.ndarray, ydata : np.ndarray)
        If input arguments are not numpy arrays, they will be converted to np.ndarray. Otherwise, the original arrays are returned.


    Raises
    ------
    ValueError
        If xdata and ydata are not the same shape
    """
    # verify shape and orientation is 1D
    xdata, ydata = verify_pair_of_1D_arrays_type_and_shape(xdata, ydata)
    # verify resulting 1D arrays are same length
    verify_pair_of_1D_arrays_lengths(xdata, ydata)
    return xdata, ydata



"""

Testing get_electron_saturation_current func - funcs_electon_saturation_current.py


"""
from typing import Tuple, Dict, Any

import numpy as np
import numpy.typing as npt
from numpy.polynomial import Polynomial as P

from pstl.utls.verify import verify_type, verify_array_polarity
from pstl.utls.functionfit.helpers import find_fit
from pstl.diagnostics.probes.langmuir.single.analysis.floating_potential import get_floating_potential
from pstl.diagnostics.probes.langmuir.single.analysis.electron_temperaure import get_electron_temperature
# from funcs_plasma_potential import get_plasma_potential

default_fit_kwargs = {
    'deg': 1, 'power': 1, 'polarity': 1,
    'reverse': True, 'return_best': True, 'fit_type': "exponential",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': 0.1, 'fstep_type': 'percent', 'fstep_adjust': True, 'fitmax': None,
    'bstep': 1, 'bstep_type': 'index', 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.30,
    'threshold_rsq': 0.95, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 1.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}


def get_electron_saturation_current(*args, method=0, **kwargs) -> Tuple[float, Dict[str, Any]]:
    # Declare available methods
    available_methods = {
        0: 'fit',
    }

    # Converts method: str -> method: int if method is a str
    if isinstance(method, str):
        reversed_methods = {v: k for k, v in available_methods.items()}
        method = reversed_methods.get(method, None)

    # check for match and get which function to use
    # raises value error with options if failed to match
    if method == 0:  # default
        func = get_electron_saturation_current_fit
    else:  # makes a table of options if error occurs
        table = "\n".join([f"{k}\t{v}" for k, v in available_methods.items()])
        raise ValueError(
            f"Matching method not found: {method}\nChoose from one of the available options:\n{table}")

    # Call funtion and return result
    return func(*args, **kwargs)


def get_electron_saturation_current_density(
        area, *args,
        method=0, I_es=None,
        **kwargs) -> Tuple[float, Dict[str, Any]]:

    # if electron saturation currint is None solve with defaults
    if I_es is None:
        value, other = get_electron_saturation_current(
            *args, method=method, **kwargs)
        I_es = value
    else:
        other = {}
    # then get electron saturatin current density
    J_es = np.divide(I_es, area)
    return J_es, other


def get_electron_saturation_current_fit(
        voltage, current, *args,
        V_f=None, V_s=None, elec_ret_poly=None, min_points=5, find_V_s=False,
        **kwargs) -> Tuple[float, Dict[str, Any]]:

    # determine starting point (all positive after vf) if not given
    if V_f is None:
        # determine starting point (all positive after V_f)
        floating_kwargs = kwargs.pop('V_f_kwargs', {})
        floating_kwargs.setdefault('method', "consecutive")
        # get floating potential
        V_f, _ = get_floating_potential(
            voltage, current, **floating_kwargs)
    # verify V_f
    verify_type(V_f, (int, float, np.int64, np.float64, np.ndarray), 'V_f')

    # make sure have Electron Retarding Region Polynomial of tpye ~np.polynomial.Polynomial
    if elec_ret_poly is None:
        KT_e_kwargs = kwargs.get("KT_e_kwargs", {})
        # set defaults for getting temperature
        KT_e_kwargs.setdefault('V_f', V_f)
        KT_e_kwargs.setdefault('V_s', V_s)
        KT_e_kwargs.setdefault('method', 'fit')
        # overide fit kwargs defaults for electron temperaure fit (exponential)
        KT_e_kwargs.setdefault('fit_kwargs',
                               {
                                   'full': False,
                               }
                               )
        print(KT_e_kwargs)
        _, KT_e_others = get_electron_temperature(
            voltage, current, **KT_e_kwargs)
        elec_ret_poly = KT_e_others["fit"].poly.convert()
    verify_type(elec_ret_poly, P)

    # use vs as a short cut if given
    if V_s is not None:
        verify_type(V_s, (int, float, np.int64, np.float64, np.ndarray), 'V_s')
        istart = np.where(voltage < V_s)[0][-1]+1
    else:  # uses vf
        istart = np.where(voltage < V_f)[0][-1]+1

    # get stoping splicing index of data (aka length of data)
    iend = len(voltage)

    # set some defaults for routine
    fit_kwargs = dict(default_fit_kwargs)
    fit_kwargs.update(kwargs.pop('fit_kwargs', {}))

    # initialiaze for search of electron saturation region
    electron_saturation_current = None
    elec_sat_fit = None
    intersection = None
    # the while loop checks to make sure the determined electron saturation fit
    # is indeed reasonable by taking the total residual by taking the differnece between fit line and
    # experimental data in the semilogy plane for all voltages between istart to iend (gets less relaiable if V_s is given.
    # The total residual should be negative as a good fit will have the electron saturation fit almost always be either much
    # greater (magnitudes difference) than experimental or almost exactly the same. However a bad fit will be the inverse
    while electron_saturation_current is None and iend >= min_points:

        # get search area data
        xdata = voltage[istart:iend]
        ydata = current[istart:iend]

        # find potential electron saturation region fit
        elec_sat_fit = find_fit(xdata, ydata, **fit_kwargs)
        # convert the fit to polynomial of type ~np.polynomial.Polynomial with satandared coefs
        elec_sat_poly = elec_sat_fit.poly.convert()

        # go from start of floating to end of electron saturation region
        # float point cuz want all the way to the end for most residual change
        # end of elec sat cuz want to not take into account the breakdown current if that occurs
        jend = np.where(xdata <= elec_sat_fit.domain[1])[0][-1]
        x = xdata[istart:jend]
        y = ydata[istart:jend]

        # covnvert to semilogy space (experimental)
        logy = np.log(y)

        # covnvert to semilogy space (fit)
        logyp = elec_sat_fit(x)

        # calculate total residual
        residuals = np.subtract(logy, logyp)
        residual = np.sum(residuals)

        # means the fit is below experimental data (may have found probe breakdown current)
        if residual > 0:
            iend -= 1
            continue
        elif residual <= 0:  # the fit is above experimental and intersection could be Vs
            # find intersection
            p = elec_sat_poly-elec_ret_poly
            intersection = p.roots()

            # Check if intersection has a value and break if it does or continue if it doesnt
            if intersection is None:
                iend -= 1
                continue
            else:
                break
        else:
            raise ValueError(
                "'residiual' is unexpected value: %s" % (str(residual)))

    # Checks that the intersection is found
    if intersection is None:
        raise ValueError("Intersection not found")
    else:
        # Determine new V_s based on intersection point
        # future work: maybe add check here that only one vs is found
        V_s = intersection[0]

    # checks that an elec_sat_fit was made
    if elec_sat_fit is None:
        raise ValueError("'elec_sat_fit' not found")
    # calculate electron saturation current
    electron_saturation_current = elec_sat_fit(V_s)

    # determine returns based on input arguments
    others: dict[str, Any] = {"fit": elec_sat_fit}
    if find_V_s:
        others["V_s"] = V_s

    # return as (value, Dict[str,Any])
    return electron_saturation_current, others



import numpy as np


m_p = 1.67262192369e-27  # kg
m_e = 9.1093837015e-31  # kg
q_e = 1.60217663e-19  # coulombs
e = 1.60217663e-19  # coulombs
q = 1.60217663e-19  # coulombs

# Boltzman Constant
K_B = 1.380649e-23  # J/K --or-- m2 kg s-2 K-1
K_B_eV = 8.617333262e-5  # eV/K

# Permitivity of free space
epsilon_0 = 8.85418782e-12  # m-3 kg-1 s4 A2


# Debye length
def lambda_D(n, KT_e):
    # T_e : eV
    return np.sqrt(
        np.divide(
            np.multiply(KT_e*e, epsilon_0),
            np.multiply(n, np.multiply(q_e, q_e))
        )
    )
